"""
Support for iptables

Configuration Options
---------------------

The following options can be set in the minion config, grains, pillar, or
master config. The configuration is read using :py:func:`config.get
<salt.modules.config.get>`.

- ``iptables.save_filters``: List of REGEX strings to FILTER OUT matching lines

  This is useful for filtering out chains, rules, etc that you do not wish to
  persist, such as ephemeral Docker rules.

  The default is to not filter out anything.

  .. code-block:: yaml

      iptables.save_filters:
        - "-j CATTLE_PREROUTING"
        - "-j DOCKER"
        - "-A POSTROUTING"
        - "-A CATTLE_POSTROUTING"
        - "-A FORWARD"
"""

import logging
import os
import re
import string
import sys
import uuid

import salt.utils.args
import salt.utils.files
import salt.utils.path
from salt.exceptions import SaltException
from salt.state import STATE_INTERNAL_KEYWORDS as _STATE_INTERNAL_KEYWORDS

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load the module if iptables is installed
    """
    if not salt.utils.path.which("iptables"):
        return (
            False,
            "The iptables execution module cannot be loaded: iptables not installed.",
        )

    return True


def _iptables_cmd(family="ipv4"):
    """
    Return correct command based on the family, e.g. ipv4 or ipv6
    """
    if family == "ipv6":
        return salt.utils.path.which("ip6tables")
    else:
        return salt.utils.path.which("iptables")


def _has_option(option, family="ipv4"):
    """
    Return truth of whether iptables has `option`.  For example:

    .. code-block:: python

        _has_option('--wait')
        _has_option('--check', family='ipv6')
    """
    cmd = "{} --help".format(_iptables_cmd(family))
    if option in __salt__["cmd.run_stdout"](cmd, output_loglevel="quiet"):
        return True
    return False


def _conf(family="ipv4"):
    """
    Some distros have a specific location for config files
    """
    if __grains__["os_family"] == "RedHat":
        if family == "ipv6":
            return "/etc/sysconfig/ip6tables"
        else:
            return "/etc/sysconfig/iptables"
    elif __grains__["os_family"] == "Arch":
        if family == "ipv6":
            return "/etc/iptables/ip6tables.rules"
        else:
            return "/etc/iptables/iptables.rules"
    elif __grains__["os_family"] == "Debian":
        if family == "ipv6":
            return "/etc/iptables/rules.v6"
        else:
            return "/etc/iptables/rules.v4"
    elif __grains__["os"] == "Gentoo":
        if family == "ipv6":
            return "/var/lib/ip6tables/rules-save"
        else:
            return "/var/lib/iptables/rules-save"
    elif __grains__["os_family"] == "Suse":
        # SuSE does not seem to use separate files for IPv4 and IPv6
        return "/etc/sysconfig/scripts/SuSEfirewall2-custom"
    elif __grains__["os_family"] == "Void":
        if family == "ipv4":
            return "/etc/iptables/iptables.rules"
        else:
            return "/etc/iptables/ip6tables.rules"
    elif __grains__["os"] == "Alpine":
        if family == "ipv6":
            return "/etc/iptables/rules6-save"
        else:
            return "/etc/iptables/rules-save"
    elif __grains__["os_family"] == "NILinuxRT":
        if family == "ipv6":
            return "/etc/natinst/share/ip6tables.conf"
        else:
            return "/etc/natinst/share/iptables.conf"
    else:
        raise SaltException(
            "Saving iptables to file is not"
            + " supported on {}.".format(__grains__["os"])
            + " Please file an issue with SaltStack"
        )


def _conf_save_filters():
    """
    Return array of strings from `save_filters` in config.

    This array will be pulled from minion config, minion grains,
    minion pillar, or master config.  The default value returned is [].

    .. code-block:: python

        _conf_save_filters()
    """
    config = __salt__["config.option"]("iptables.save_filters", [])
    return config


def _regex_iptables_save(cmd_output, filters=None):
    """
    Return string with `save_filter` regex entries removed.  For example:

    If `filters` is not provided, it will be pulled from minion config,
    minion grains, minion pillar, or master config. Default return value
    if no filters found is the original cmd_output string.

    .. code-block:: python

        _regex_iptables_save(cmd_output, ['-A DOCKER*'])
    """
    # grab RE compiled filters from context for performance
    if "iptables.save_filters" not in __context__:
        __context__["iptables.save_filters"] = []
        for pattern in filters or _conf_save_filters():
            try:
                __context__["iptables.save_filters"].append(re.compile(pattern))
            except re.error as e:
                log.warning("Skipping regex rule: '%s': %s", pattern, e)
                continue

    if __context__["iptables.save_filters"]:
        # line by line get rid of any regex matches
        _filtered_cmd_output = [
            line
            for line in cmd_output.splitlines(True)
            if not any(reg.search(line) for reg in __context__["iptables.save_filters"])
        ]
        return "".join(_filtered_cmd_output)

    return cmd_output


def version(family="ipv4"):
    """
    Return version from iptables --version

    CLI Example:

    .. code-block:: bash

        salt '*' iptables.version

        IPv6:
        salt '*' iptables.version family=ipv6
    """
    cmd = "{} --version".format(_iptables_cmd(family))
    out = __salt__["cmd.run_stdout"](cmd).split()
    return out[1]


def build_rule(
    table="filter",
    chain=None,
    command=None,
    position="",
    full=None,
    family="ipv4",
    **kwargs
):
    """
    Build a well-formatted iptables rule based on kwargs. A `table` and `chain`
    are not required, unless `full` is True.

    If `full` is `True`, then `table`, `chain` and `command` are required.
    `command` may be specified as either a short option ('I') or a long option
    (`--insert`). This will return the iptables command, exactly as it would
    be used from the command line.

    If a position is required (as with `-I` or `-D`), it may be specified as
    `position`. This will only be useful if `full` is True.

    If `state` is passed, it will be ignored, use `connstate`.
    If `connstate` is passed in, it will automatically be changed to `state`.

    To pass in jump options that doesn't take arguments, pass in an empty
    string.

    .. note::

        Whereas iptables will accept ``-p``, ``--proto[c[o[l]]]`` as synonyms
        of ``--protocol``, if ``--proto`` appears in an iptables command after
        the appearance of ``-m policy``, it is interpreted as the ``--proto``
        option of the policy extension (see the iptables-extensions(8) man
        page).

    CLI Examples:

    .. code-block:: bash

        salt '*' iptables.build_rule match=state \\
            connstate=RELATED,ESTABLISHED jump=ACCEPT

        salt '*' iptables.build_rule filter INPUT command=I position=3 \\
            full=True match=state connstate=RELATED,ESTABLISHED jump=ACCEPT

        salt '*' iptables.build_rule filter INPUT command=A \\
            full=True match=state connstate=RELATED,ESTABLISHED \\
            source='127.0.0.1' jump=ACCEPT

        .. Invert Rules
        salt '*' iptables.build_rule filter INPUT command=A \\
            full=True match=state connstate=RELATED,ESTABLISHED \\
            source='!127.0.0.1' jump=ACCEPT

        salt '*' iptables.build_rule filter INPUT command=A \\
            full=True match=state connstate=RELATED,ESTABLISHED \\
            destination='not 127.0.0.1' jump=ACCEPT

        IPv6:
        salt '*' iptables.build_rule match=state \\
            connstate=RELATED,ESTABLISHED jump=ACCEPT \\
            family=ipv6
        salt '*' iptables.build_rule filter INPUT command=I position=3 \\
            full=True match=state connstate=RELATED,ESTABLISHED jump=ACCEPT \\
            family=ipv6
    """
    if "target" in kwargs:
        kwargs["jump"] = kwargs.pop("target")

    # Ignore name and state for this function
    kwargs.pop("name", None)
    kwargs.pop("state", None)

    for ignore in list(_STATE_INTERNAL_KEYWORDS) + ["chain", "save", "table"]:
        if ignore in kwargs:
            del kwargs[ignore]

    rule = []
    protocol = False
    bang_not_pat = re.compile(r"(!|not)\s?")

    def maybe_add_negation(arg):
        """
        Will check if the defined argument is intended to be negated,
        (i.e. prefixed with '!' or 'not'), and add a '! ' to the rule.

        The prefix will be removed from the value in the kwargs dict.
        """
        value = str(kwargs[arg])
        if value.startswith("!") or value.startswith("not"):
            kwargs[arg] = re.sub(bang_not_pat, "", value)
            return "! "
        return ""

    if "if" in kwargs:
        rule.append("{}-i {}".format(maybe_add_negation("if"), kwargs["if"]))
        del kwargs["if"]

    if "of" in kwargs:
        rule.append("{}-o {}".format(maybe_add_negation("of"), kwargs["of"]))
        del kwargs["of"]

    if "proto" in kwargs and kwargs.get("match") != "policy":
        kwargs["protocol"] = kwargs["proto"]
        del kwargs["proto"]
        # Handle the case 'proto' in kwargs and kwargs.get('match') == 'policy' below
    if "protocol" in kwargs:
        if not protocol:
            rule.append(
                "{}-p {}".format(maybe_add_negation("protocol"), kwargs["protocol"])
            )
            protocol = True
        del kwargs["protocol"]

    if "match" in kwargs:
        match_value = kwargs["match"]
        if not isinstance(match_value, list):
            match_value = match_value.split(",")
        for match in match_value:
            rule.append("-m {}".format(match))
            if "name_" in kwargs and match.strip() in ("pknock", "quota2", "recent"):
                rule.append("--name {}".format(kwargs["name_"]))
                del kwargs["name_"]
        if "proto" in kwargs and kwargs.get("match") == "policy":
            rule.append(
                "{}--proto {}".format(maybe_add_negation("proto"), kwargs["proto"])
            )
            del kwargs["proto"]
        del kwargs["match"]

    if "match-set" in kwargs:
        if isinstance(kwargs["match-set"], str):
            kwargs["match-set"] = [kwargs["match-set"]]
        for match_set in kwargs["match-set"]:
            negative_match_set = ""
            if match_set.startswith("!") or match_set.startswith("not"):
                negative_match_set = "! "
                match_set = re.sub(bang_not_pat, "", match_set)
            rule.append("-m set {}--match-set {}".format(negative_match_set, match_set))
        del kwargs["match-set"]

    if "connstate" in kwargs:
        if "-m state" not in rule:
            rule.append("-m state")

        rule.append(
            "{}--state {}".format(maybe_add_negation("connstate"), kwargs["connstate"])
        )

        del kwargs["connstate"]

    if "dport" in kwargs:
        rule.append("{}--dport {}".format(maybe_add_negation("dport"), kwargs["dport"]))
        del kwargs["dport"]

    if "sport" in kwargs:
        rule.append("{}--sport {}".format(maybe_add_negation("sport"), kwargs["sport"]))
        del kwargs["sport"]

    for multiport_arg in ("dports", "sports"):
        if multiport_arg in kwargs:
            if "-m multiport" not in rule:
                rule.append("-m multiport")
                if not protocol:
                    return "Error: protocol must be specified"

            mp_value = kwargs[multiport_arg]
            if isinstance(mp_value, list):
                if any(
                    i
                    for i in mp_value
                    if str(i).startswith("!") or str(i).startswith("not")
                ):
                    mp_value = [
                        re.sub(bang_not_pat, "", str(item)) for item in mp_value
                    ]
                    rule.append("!")
                dports = ",".join(str(i) for i in mp_value)
            else:
                if str(mp_value).startswith("!") or str(mp_value).startswith("not"):
                    dports = re.sub(bang_not_pat, "", mp_value)
                    rule.append("!")
                else:
                    dports = mp_value

            rule.append("--{} {}".format(multiport_arg, dports))
            del kwargs[multiport_arg]

    if "comment" in kwargs:
        if "-m comment" not in rule:
            rule.append("-m comment")

        rule.append('--comment "{}"'.format(kwargs["comment"]))
        del kwargs["comment"]

    # --set in ipset is deprecated, works but returns error.
    # rewrite to --match-set if not empty, otherwise treat as recent option
    if "set" in kwargs and kwargs["set"]:
        rule.append("{}--match-set {}".format(maybe_add_negation("set"), kwargs["set"]))
        del kwargs["set"]

    # Jumps should appear last, except for any arguments that are passed to
    # jumps, which of course need to follow.
    after_jump = []
    # All jump arguments as extracted from man iptables-extensions, man iptables,
    # man xtables-addons and http://www.iptables.info/en/iptables-targets-and-jumps.html
    after_jump_arguments = (
        "j",  # j and jump needs to be first
        "jump",
        # IPTABLES
        "add-set",
        "and-mark",
        "and-tos",
        "checksum-fill",
        "clamp-mss-to-pmtu",
        "clustermac",
        "ctevents",
        "ctmask",
        "del-set",
        "ecn-tcp-remove",
        "exist",
        "expevents",
        "gateway",
        "hash-init",
        "hashmode",
        "helper",
        "label",
        "local-node",
        "log-ip-options",
        "log-level",
        "log-prefix",
        "log-tcp-options",
        "log-tcp-sequence",
        "log-uid",
        "mask",
        "new",
        "nfmask",
        "nflog-group",
        "nflog-prefix",
        "nflog-range",
        "nflog-threshold",
        "nodst",
        "notrack",
        "on-ip",
        "on-port",
        "or-mark",
        "or-tos",
        "persistent",
        "queue-balance",
        "queue-bypass",
        "queue-num",
        "random",
        "rateest-ewmalog",
        "rateest-interval",
        "rateest-name",
        "reject-with",
        "restore",
        "restore-mark",
        #'save',  # no arg, problematic name: How do we avoid collision with this?
        "save-mark",
        "selctx",
        "set-class",
        "set-dscp",
        "set-dscp-class",
        "set-mark",
        "set-mss",
        "set-tos",
        "set-xmark",
        "strip-options",
        "timeout",
        "to",
        "to-destination",
        "to-ports",
        "to-source",
        "total-nodes",
        "tproxy-mark",
        "ttl-dec",
        "ttl-inc",
        "ttl-set",
        "type",
        "ulog-cprange",
        "ulog-nlgroup",
        "ulog-prefix",
        "ulog-qthreshold",
        "xor-mark",
        "xor-tos",
        "zone",
        # IPTABLES-EXTENSIONS
        "dst-pfx",
        "hl-dec",
        "hl-inc",
        "hl-set",
        "hmark-dport-mask",
        "hmark-dst-prefix",
        "hmark-mod",
        "hmark-offset",
        "hmark-proto-mask",
        "hmark-rnd",
        "hmark-spi-mask",
        "hmark-sport-mask",
        "hmark-src-prefix",
        "hmark-tuple",
        "led-always-blink",
        "led-delay",
        "led-trigger-id",
        "queue-cpu-fanout",
        "src-pfx",
        # WEB
        "to-port",
        # XTABLES
        "addr",
        "and-mask",
        "delude",
        "honeypot",
        "or-mask",
        "prefix",
        "reset",
        "reuse",
        "set-mac",
        "shift",
        "static",
        "tarpit",
        "tname",
        "ttl",
    )
    for after_jump_argument in after_jump_arguments:
        if after_jump_argument in kwargs:
            value = kwargs[after_jump_argument]
            if value in (None, ""):  # options without arguments
                after_jump.append("--{}".format(after_jump_argument))
            elif any(ws_char in str(value) for ws_char in string.whitespace):
                after_jump.append('--{} "{}"'.format(after_jump_argument, value))
            else:
                after_jump.append("--{} {}".format(after_jump_argument, value))
            del kwargs[after_jump_argument]

    for key in kwargs:
        negation = maybe_add_negation(key)
        # don't use .items() since maybe_add_negation removes the prefix from
        # the value in the kwargs, thus we need to fetch it after that has run
        value = kwargs[key]
        flag = "-" if len(key) == 1 else "--"
        value = "" if value in (None, "") else " {}".format(value)
        rule.append("{}{}{}{}".format(negation, flag, key, value))

    rule += after_jump

    if full:
        if not table:
            return "Error: Table needs to be specified"
        if not chain:
            return "Error: Chain needs to be specified"
        if not command:
            return "Error: Command needs to be specified"

        if command in "ACDIRLSFZNXPE":
            flag = "-"
        else:
            flag = "--"

        wait = "--wait" if _has_option("--wait", family) else ""

        return "{} {} -t {} {}{} {} {} {}".format(
            _iptables_cmd(family),
            wait,
            table,
            flag,
            command,
            chain,
            position,
            " ".join(rule),
        )

    return " ".join(rule)


def get_saved_rules(conf_file=None, family="ipv4"):
    """
    Return a data structure of the rules in the conf file

    CLI Example:

    .. code-block:: bash

        salt '*' iptables.get_saved_rules

        IPv6:
        salt '*' iptables.get_saved_rules family=ipv6
    """
    return _parse_conf(conf_file=conf_file, family=family)


def get_rules(family="ipv4"):
    """
    Return a data structure of the current, in-memory rules

    CLI Example:

    .. code-block:: bash

        salt '*' iptables.get_rules

        IPv6:
        salt '*' iptables.get_rules family=ipv6

    """
    return _parse_conf(in_mem=True, family=family)


def get_saved_policy(table="filter", chain=None, conf_file=None, family="ipv4"):
    """
    Return the current policy for the specified table/chain

    CLI Examples:

    .. code-block:: bash

        salt '*' iptables.get_saved_policy filter INPUT
        salt '*' iptables.get_saved_policy filter INPUT \\
            conf_file=/etc/iptables.saved

        IPv6:
        salt '*' iptables.get_saved_policy filter INPUT family=ipv6
        salt '*' iptables.get_saved_policy filter INPUT \\
            conf_file=/etc/iptables.saved family=ipv6

    """
    if not chain:
        return "Error: Chain needs to be specified"

    rules = _parse_conf(conf_file, family=family)
    try:
        return rules[table][chain]["policy"]
    except KeyError:
        return None


def get_policy(table="filter", chain=None, family="ipv4"):
    """
    Return the current policy for the specified table/chain

    CLI Example:

    .. code-block:: bash

        salt '*' iptables.get_policy filter INPUT

        IPv6:
        salt '*' iptables.get_policy filter INPUT family=ipv6
    """
    if not chain:
        return "Error: Chain needs to be specified"

    rules = _parse_conf(in_mem=True, family=family)
    try:
        return rules[table][chain]["policy"]
    except KeyError:
        return None


def set_policy(table="filter", chain=None, policy=None, family="ipv4"):
    """
    Set the current policy for the specified table/chain

    CLI Example:

    .. code-block:: bash

        salt '*' iptables.set_policy filter INPUT ACCEPT

        IPv6:
        salt '*' iptables.set_policy filter INPUT ACCEPT family=ipv6
    """
    if not chain:
        return "Error: Chain needs to be specified"
    if not policy:
        return "Error: Policy needs to be specified"

    wait = "--wait" if _has_option("--wait", family) else ""
    cmd = "{} {} -t {} -P {} {}".format(
        _iptables_cmd(family), wait, table, chain, policy
    )
    out = __salt__["cmd.run_stderr"](cmd)
    return out


def save(filename=None, family="ipv4"):
    """
    Save the current in-memory rules to disk

    CLI Example:

    .. code-block:: bash

        salt '*' iptables.save /etc/sysconfig/iptables

        IPv6:
        salt '*' iptables.save /etc/sysconfig/iptables family=ipv6
    """
    if _conf() and not filename:
        filename = _conf(family)

    log.debug("Saving rules to %s", filename)

    parent_dir = os.path.dirname(filename)
    if not os.path.isdir(parent_dir):
        os.makedirs(parent_dir)
    cmd = "{}-save".format(_iptables_cmd(family))
    ipt = __salt__["cmd.run_stdout"](cmd)

    # regex out the output if configured with filters
    if _conf_save_filters():
        ipt = _regex_iptables_save(ipt)

    out = __salt__["file.write"](filename, ipt)
    return out


def check(table="filter", chain=None, rule=None, family="ipv4"):
    """
    Check for the existence of a rule in the table and chain

    This function accepts a rule in a standard iptables command format,
        starting with the chain. Trying to force users to adapt to a new
        method of creating rules would be irritating at best, and we
        already have a parser that can handle it.

    CLI Example:

    .. code-block:: bash

        salt '*' iptables.check filter INPUT \\
            rule='-m state --state RELATED,ESTABLISHED -j ACCEPT'

        IPv6:
        salt '*' iptables.check filter INPUT \\
            rule='-m state --state RELATED,ESTABLISHED -j ACCEPT' \\
            family=ipv6
    """
    if not chain:
        return "Error: Chain needs to be specified"
    if not rule:
        return "Error: Rule needs to be specified"
    ipt_cmd = _iptables_cmd(family)

    if _has_option("--check", family):
        cmd = "{} -t {} -C {} {}".format(ipt_cmd, table, chain, rule)
        out = __salt__["cmd.run_stderr"](cmd, output_loglevel="quiet")
    else:
        _chain_name = hex(uuid.getnode())

        # Create temporary table
        __salt__["cmd.run"]("{} -t {} -N {}".format(ipt_cmd, table, _chain_name))
        __salt__["cmd.run"](
            "{} -t {} -A {} {}".format(ipt_cmd, table, _chain_name, rule)
        )

        out = __salt__["cmd.run_stdout"]("{}-save".format(ipt_cmd))

        # Clean up temporary table
        __salt__["cmd.run"]("{} -t {} -F {}".format(ipt_cmd, table, _chain_name))
        __salt__["cmd.run"]("{} -t {} -X {}".format(ipt_cmd, table, _chain_name))

        for i in out.splitlines():
            if i.startswith("-A {}".format(_chain_name)):
                if i.replace(_chain_name, chain) in out.splitlines():
                    return True

        return False

    if not out:
        return True
    return out


def check_chain(table="filter", chain=None, family="ipv4"):
    """
    .. versionadded:: 2014.1.0

    Check for the existence of a chain in the table

    CLI Example:

    .. code-block:: bash

        salt '*' iptables.check_chain filter INPUT

        IPv6:
        salt '*' iptables.check_chain filter INPUT family=ipv6
    """

    if not chain:
        return "Error: Chain needs to be specified"

    cmd = "{}-save -t {}".format(_iptables_cmd(family), table)
    out = __salt__["cmd.run_stdout"](cmd).find(":{} ".format(chain))

    if out != -1:
        out = True
    else:
        out = False

    return out


def new_chain(table="filter", chain=None, family="ipv4"):
    """
    .. versionadded:: 2014.1.0

    Create new custom chain to the specified table.

    CLI Example:

    .. code-block:: bash

        salt '*' iptables.new_chain filter CUSTOM_CHAIN

        IPv6:
        salt '*' iptables.new_chain filter CUSTOM_CHAIN family=ipv6
    """

    if not chain:
        return "Error: Chain needs to be specified"

    wait = "--wait" if _has_option("--wait", family) else ""
    cmd = "{} {} -t {} -N {}".format(_iptables_cmd(family), wait, table, chain)
    out = __salt__["cmd.run_stderr"](cmd)

    if not out:
        out = True
    return out


def delete_chain(table="filter", chain=None, family="ipv4"):
    """
    .. versionadded:: 2014.1.0

    Delete custom chain to the specified table.

    CLI Example:

    .. code-block:: bash

        salt '*' iptables.delete_chain filter CUSTOM_CHAIN

        IPv6:
        salt '*' iptables.delete_chain filter CUSTOM_CHAIN family=ipv6
    """

    if not chain:
        return "Error: Chain needs to be specified"

    wait = "--wait" if _has_option("--wait", family) else ""
    cmd = "{} {} -t {} -X {}".format(_iptables_cmd(family), wait, table, chain)
    out = __salt__["cmd.run_stderr"](cmd)

    if not out:
        out = True
    return out


def append(table="filter", chain=None, rule=None, family="ipv4"):
    """
    Append a rule to the specified table/chain.

    This function accepts a rule in a standard iptables command format,
        starting with the chain. Trying to force users to adapt to a new
        method of creating rules would be irritating at best, and we
        already have a parser that can handle it.

    CLI Example:

    .. code-block:: bash

        salt '*' iptables.append filter INPUT \\
            rule='-m state --state RELATED,ESTABLISHED -j ACCEPT'

        IPv6:
        salt '*' iptables.append filter INPUT \\
            rule='-m state --state RELATED,ESTABLISHED -j ACCEPT' \\
            family=ipv6
    """
    if not chain:
        return "Error: Chain needs to be specified"
    if not rule:
        return "Error: Rule needs to be specified"

    wait = "--wait" if _has_option("--wait", family) else ""
    returnCheck = check(table, chain, rule, family)
    if isinstance(returnCheck, bool) and returnCheck:
        return False
    cmd = "{} {} -t {} -A {} {}".format(_iptables_cmd(family), wait, table, chain, rule)
    out = __salt__["cmd.run_stderr"](cmd)
    return not out


def insert(table="filter", chain=None, position=None, rule=None, family="ipv4"):
    """
    Insert a rule into the specified table/chain, at the specified position.

    This function accepts a rule in a standard iptables command format,
        starting with the chain. Trying to force users to adapt to a new
        method of creating rules would be irritating at best, and we
        already have a parser that can handle it.

    If the position specified is a negative number, then the insert will be
        performed counting from the end of the list. For instance, a position
        of -1 will insert the rule as the second to last rule. To insert a rule
        in the last position, use the append function instead.

    CLI Examples:

    .. code-block:: bash

        salt '*' iptables.insert filter INPUT position=3 \\
            rule='-m state --state RELATED,ESTABLISHED -j ACCEPT'

        IPv6:
        salt '*' iptables.insert filter INPUT position=3 \\
            rule='-m state --state RELATED,ESTABLISHED -j ACCEPT' \\
            family=ipv6
    """
    if not chain:
        return "Error: Chain needs to be specified"
    if not position:
        return "Error: Position needs to be specified or use append (-A)"
    if not rule:
        return "Error: Rule needs to be specified"

    if position < 0:
        rules = get_rules(family=family)
        size = len(rules[table][chain]["rules"])
        position = (size + position) + 1
        if position == 0:
            position = 1

    wait = "--wait" if _has_option("--wait", family) else ""
    returnCheck = check(table, chain, rule, family)
    if isinstance(returnCheck, bool) and returnCheck:
        return False
    cmd = "{} {} -t {} -I {} {} {}".format(
        _iptables_cmd(family), wait, table, chain, position, rule
    )
    out = __salt__["cmd.run_stderr"](cmd)
    return out


def delete(table, chain=None, position=None, rule=None, family="ipv4"):
    """
    Delete a rule from the specified table/chain, specifying either the rule
        in its entirety, or the rule's position in the chain.

    This function accepts a rule in a standard iptables command format,
        starting with the chain. Trying to force users to adapt to a new
        method of creating rules would be irritating at best, and we
        already have a parser that can handle it.

    CLI Examples:

    .. code-block:: bash

        salt '*' iptables.delete filter INPUT position=3
        salt '*' iptables.delete filter INPUT \\
            rule='-m state --state RELATED,ESTABLISHED -j ACCEPT'

        IPv6:
        salt '*' iptables.delete filter INPUT position=3 family=ipv6
        salt '*' iptables.delete filter INPUT \\
            rule='-m state --state RELATED,ESTABLISHED -j ACCEPT' \\
            family=ipv6
    """

    if position and rule:
        return "Error: Only specify a position or a rule, not both"

    if position:
        rule = position

    wait = "--wait" if _has_option("--wait", family) else ""
    cmd = "{} {} -t {} -D {} {}".format(_iptables_cmd(family), wait, table, chain, rule)
    out = __salt__["cmd.run_stderr"](cmd)
    return out


def flush(table="filter", chain="", family="ipv4"):
    """
    Flush the chain in the specified table, flush all chains in the specified
    table if not specified chain.

    CLI Example:

    .. code-block:: bash

        salt '*' iptables.flush filter INPUT

        IPv6:
        salt '*' iptables.flush filter INPUT family=ipv6
    """

    wait = "--wait" if _has_option("--wait", family) else ""
    cmd = "{} {} -t {} -F {}".format(_iptables_cmd(family), wait, table, chain)
    out = __salt__["cmd.run_stderr"](cmd)
    return out


def _parse_conf(conf_file=None, in_mem=False, family="ipv4"):
    """
    If a file is not passed in, and the correct one for this OS is not
    detected, return False
    """
    if _conf() and not conf_file and not in_mem:
        conf_file = _conf(family)

    rules = ""
    if conf_file:
        with salt.utils.files.fopen(conf_file, "r") as ifile:
            rules = ifile.read()
    elif in_mem:
        cmd = "{}-save".format(_iptables_cmd(family))
        rules = __salt__["cmd.run_stdout"](cmd)
    else:
        raise SaltException("A file was not found to parse")

    ret = {}
    table = ""
    parser = _parser()
    for line in rules.splitlines():
        line = salt.utils.stringutils.to_unicode(line)
        if line.startswith("*"):
            table = line.replace("*", "")
            ret[table] = {}
        elif line.startswith(":"):
            comps = line.split()
            chain = comps[0].replace(":", "")
            ret[table][chain] = {}
            ret[table][chain]["policy"] = comps[1]
            counters = comps[2].replace("[", "").replace("]", "")
            (pcount, bcount) = counters.split(":")
            ret[table][chain]["packet count"] = pcount
            ret[table][chain]["byte count"] = bcount
            ret[table][chain]["rules"] = []
            ret[table][chain]["rules_comment"] = {}
        elif line.startswith("-A"):
            args = salt.utils.args.shlex_split(line)
            index = 0
            while index + 1 < len(args):
                swap = args[index] == "!" and args[index + 1].startswith("-")
                if swap:
                    args[index], args[index + 1] = args[index + 1], args[index]
                if args[index].startswith("-"):
                    index += 1
                    if args[index].startswith("-") or (args[index] == "!" and not swap):
                        args.insert(index, "")
                    else:
                        while (
                            index + 1 < len(args)
                            and args[index + 1] != "!"
                            and not args[index + 1].startswith("-")
                        ):
                            args[index] += " {}".format(args.pop(index + 1))
                index += 1
            if args[-1].startswith("-"):
                args.append("")
            parsed_args = []
            opts, _ = parser.parse_known_args(args)
            parsed_args = vars(opts)
            ret_args = {}
            chain = parsed_args["append"]
            for arg in parsed_args:
                if parsed_args[arg] and arg != "append":
                    ret_args[arg] = parsed_args[arg]
            if parsed_args["comment"] is not None:
                comment = parsed_args["comment"][0].strip('"')
                ret[table][chain[0]]["rules_comment"][comment] = ret_args
            ret[table][chain[0]]["rules"].append(ret_args)
    return ret


def _parser():
    """
    This function attempts to list all the options documented in the
    iptables(8) and iptables-extensions(8) man pages.  They will not all be
    used by all parts of the module; use them intelligently and appropriately.
    """
    add_arg = None
    if sys.version.startswith("2.6"):
        import optparse

        parser = optparse.OptionParser()
        add_arg = parser.add_option
    else:
        import argparse  # pylint: disable=minimum-python-version

        parser = argparse.ArgumentParser()
        add_arg = parser.add_argument

    # COMMANDS
    add_arg("-A", "--append", dest="append", action="append")
    add_arg("-D", "--delete", dest="delete", action="append")
    add_arg("-I", "--insert", dest="insert", action="append")
    add_arg("-R", "--replace", dest="replace", action="append")
    add_arg("-L", "--list", dest="list", action="append")
    add_arg("-F", "--flush", dest="flush", action="append")
    add_arg("-Z", "--zero", dest="zero", action="append")
    add_arg("-N", "--new-chain", dest="new-chain", action="append")
    add_arg("-X", "--delete-chain", dest="delete-chain", action="append")
    add_arg("-P", "--policy", dest="policy", action="append")
    add_arg("-E", "--rename-chain", dest="rename-chain", action="append")

    # PARAMETERS
    add_arg("-p", "--protocol", dest="protocol", action="append")
    add_arg("-s", "--source", dest="source", action="append")
    add_arg("-d", "--destination", dest="destination", action="append")
    add_arg("-j", "--jump", dest="jump", action="append")
    add_arg("-g", "--goto", dest="goto", action="append")
    add_arg("-i", "--in-interface", dest="in-interface", action="append")
    add_arg("-o", "--out-interface", dest="out-interface", action="append")
    add_arg("-f", "--fragment", dest="fragment", action="append")
    add_arg("-c", "--set-counters", dest="set-counters", action="append")

    # MATCH EXTENSIONS
    add_arg("-m", "--match", dest="match", action="append")
    ## addrtype
    add_arg("--src-type", dest="src-type", action="append")
    add_arg("--dst-type", dest="dst-type", action="append")
    add_arg("--limit-iface-in", dest="limit-iface-in", action="append")
    add_arg("--limit-iface-out", dest="limit-iface-out", action="append")
    ## ah
    add_arg("--ahspi", dest="ahspi", action="append")
    add_arg("--ahlen", dest="ahlen", action="append")
    add_arg("--ahres", dest="ahres", action="append")
    ## bpf
    add_arg("--bytecode", dest="bytecode", action="append")
    ## cgroup
    add_arg("--cgroup", dest="cgroup", action="append")
    ## cluster
    add_arg("--cluster-total-nodes", dest="cluster-total-nodes", action="append")
    add_arg("--cluster-local-node", dest="cluster-local-node", action="append")
    add_arg("--cluster-local-nodemask", dest="cluster-local-nodemask", action="append")
    add_arg("--cluster-hash-seed", dest="cluster-hash-seed", action="append")
    add_arg("--h-length", dest="h-length", action="append")
    add_arg("--mangle-mac-s", dest="mangle-mac-s", action="append")
    add_arg("--mangle-mac-d", dest="mangle-mac-d", action="append")
    ## comment
    add_arg("--comment", dest="comment", action="append")
    ## connbytes
    add_arg("--connbytes", dest="connbytes", action="append")
    add_arg("--connbytes-dir", dest="connbytes-dir", action="append")
    add_arg("--connbytes-mode", dest="connbytes-mode", action="append")
    ## connlabel
    add_arg("--label", dest="label", action="append")
    ## connlimit
    add_arg("--connlimit-upto", dest="connlimit-upto", action="append")
    add_arg("--connlimit-above", dest="connlimit-above", action="append")
    add_arg("--connlimit-mask", dest="connlimit-mask", action="append")
    add_arg("--connlimit-saddr", dest="connlimit-saddr", action="append")
    add_arg("--connlimit-daddr", dest="connlimit-daddr", action="append")
    ## connmark
    add_arg("--mark", dest="mark", action="append")
    ## conntrack
    add_arg("--ctstate", dest="ctstate", action="append")
    add_arg("--ctproto", dest="ctproto", action="append")
    add_arg("--ctorigsrc", dest="ctorigsrc", action="append")
    add_arg("--ctorigdst", dest="ctorigdst", action="append")
    add_arg("--ctreplsrc", dest="ctreplsrc", action="append")
    add_arg("--ctrepldst", dest="ctrepldst", action="append")
    add_arg("--ctorigsrcport", dest="ctorigsrcport", action="append")
    add_arg("--ctorigdstport", dest="ctorigdstport", action="append")
    add_arg("--ctreplsrcport", dest="ctreplsrcport", action="append")
    add_arg("--ctrepldstport", dest="ctrepldstport", action="append")
    add_arg("--ctstatus", dest="ctstatus", action="append")
    add_arg("--ctexpire", dest="ctexpire", action="append")
    add_arg("--ctdir", dest="ctdir", action="append")
    ## cpu
    add_arg("--cpu", dest="cpu", action="append")
    ## dccp
    add_arg("--sport", "--source-port", dest="source_port", action="append")
    add_arg("--dport", "--destination-port", dest="destination_port", action="append")
    add_arg("--dccp-types", dest="dccp-types", action="append")
    add_arg("--dccp-option", dest="dccp-option", action="append")
    ## devgroup
    add_arg("--src-group", dest="src-group", action="append")
    add_arg("--dst-group", dest="dst-group", action="append")
    ## dscp
    add_arg("--dscp", dest="dscp", action="append")
    add_arg("--dscp-class", dest="dscp-class", action="append")
    ## dst
    add_arg("--dst-len", dest="dst-len", action="append")
    add_arg("--dst-opts", dest="dst-opts", action="append")
    ## ecn
    add_arg("--ecn-tcp-cwr", dest="ecn-tcp-cwr", action="append")
    add_arg("--ecn-tcp-ece", dest="ecn-tcp-ece", action="append")
    add_arg("--ecn-ip-ect", dest="ecn-ip-ect", action="append")
    ## esp
    add_arg("--espspi", dest="espspi", action="append")
    ## frag
    add_arg("--fragid", dest="fragid", action="append")
    add_arg("--fraglen", dest="fraglen", action="append")
    add_arg("--fragres", dest="fragres", action="append")
    add_arg("--fragfirst", dest="fragfirst", action="append")
    add_arg("--fragmore", dest="fragmore", action="append")
    add_arg("--fraglast", dest="fraglast", action="append")
    ## hashlimit
    add_arg("--hashlimit-upto", dest="hashlimit-upto", action="append")
    add_arg("--hashlimit-above", dest="hashlimit-above", action="append")
    add_arg("--hashlimit-burst", dest="hashlimit-burst", action="append")
    add_arg("--hashlimit-mode", dest="hashlimit-mode", action="append")
    add_arg("--hashlimit-srcmask", dest="hashlimit-srcmask", action="append")
    add_arg("--hashlimit-dstmask", dest="hashlimit-dstmask", action="append")
    add_arg("--hashlimit-name", dest="hashlimit-name", action="append")
    add_arg("--hashlimit-htable-size", dest="hashlimit-htable-size", action="append")
    add_arg("--hashlimit-htable-max", dest="hashlimit-htable-max", action="append")
    add_arg(
        "--hashlimit-htable-expire", dest="hashlimit-htable-expire", action="append"
    )
    add_arg(
        "--hashlimit-htable-gcinterval",
        dest="hashlimit-htable-gcinterval",
        action="append",
    )
    ## hbh
    add_arg("--hbh-len", dest="hbh-len", action="append")
    add_arg("--hbh-opts", dest="hbh-opts", action="append")
    ## helper
    add_arg("--helper", dest="helper", action="append")
    ## hl
    add_arg("--hl-eq", dest="hl-eq", action="append")
    add_arg("--hl-lt", dest="hl-lt", action="append")
    add_arg("--hl-gt", dest="hl-gt", action="append")
    ## icmp
    add_arg("--icmp-type", dest="icmp-type", action="append")
    ## icmp6
    add_arg("--icmpv6-type", dest="icmpv6-type", action="append")
    ## iprange
    add_arg("--src-range", dest="src-range", action="append")
    add_arg("--dst-range", dest="dst-range", action="append")
    ## ipv6header
    add_arg("--soft", dest="soft", action="append")
    add_arg("--header", dest="header", action="append")
    ## ipvs
    add_arg("--ipvs", dest="ipvs", action="append")
    add_arg("--vproto", dest="vproto", action="append")
    add_arg("--vaddr", dest="vaddr", action="append")
    add_arg("--vport", dest="vport", action="append")
    add_arg("--vdir", dest="vdir", action="append")
    add_arg("--vmethod", dest="vmethod", action="append")
    add_arg("--vportctl", dest="vportctl", action="append")
    ## length
    add_arg("--length", dest="length", action="append")
    ## limit
    add_arg("--limit", dest="limit", action="append")
    add_arg("--limit-burst", dest="limit-burst", action="append")
    ## mac
    add_arg("--mac-source", dest="mac-source", action="append")
    ## mh
    add_arg("--mh-type", dest="mh-type", action="append")
    ## multiport
    add_arg("--sports", "--source-ports", dest="source-ports", action="append")
    add_arg(
        "--dports", "--destination-ports", dest="destination-ports", action="append"
    )
    add_arg("--ports", dest="ports", action="append")
    ## nfacct
    add_arg("--nfacct-name", dest="nfacct-name", action="append")
    ## osf
    add_arg("--genre", dest="genre", action="append")
    add_arg("--ttl", dest="ttl", action="append")
    add_arg("--log", dest="log", action="append")
    ## owner
    add_arg("--uid-owner", dest="uid-owner", action="append")
    add_arg("--gid-owner", dest="gid-owner", action="append")
    add_arg("--socket-exists", dest="socket-exists", action="append")
    ## physdev
    add_arg("--physdev-in", dest="physdev-in", action="append")
    add_arg("--physdev-out", dest="physdev-out", action="append")
    add_arg("--physdev-is-in", dest="physdev-is-in", action="append")
    add_arg("--physdev-is-out", dest="physdev-is-out", action="append")
    add_arg("--physdev-is-bridged", dest="physdev-is-bridged", action="append")
    ## pkttype
    add_arg("--pkt-type", dest="pkt-type", action="append")
    ## policy
    add_arg("--dir", dest="dir", action="append")
    add_arg("--pol", dest="pol", action="append")
    add_arg("--strict", dest="strict", action="append")
    add_arg("--reqid", dest="reqid", action="append")
    add_arg("--spi", dest="spi", action="append")
    add_arg("--proto", dest="proto", action="append")
    add_arg("--mode", dest="mode", action="append")
    add_arg("--tunnel-src", dest="tunnel-src", action="append")
    add_arg("--tunnel-dst", dest="tunnel-dst", action="append")
    add_arg("--next", dest="next", action="append")
    ## quota
    add_arg("--quota", dest="quota", action="append")
    ## rateest
    add_arg("--rateest", dest="rateest", action="append")
    add_arg("--rateest1", dest="rateest1", action="append")
    add_arg("--rateest2", dest="rateest2", action="append")
    add_arg("--rateest-delta", dest="rateest-delta", action="append")
    add_arg("--rateest-bps", dest="rateest-bps", action="append")
    add_arg("--rateest-bps1", dest="rateest-bps1", action="append")
    add_arg("--rateest-bps2", dest="rateest-bps2", action="append")
    add_arg("--rateest-pps", dest="rateest-pps", action="append")
    add_arg("--rateest-pps1", dest="rateest-pps1", action="append")
    add_arg("--rateest-pps2", dest="rateest-pps2", action="append")
    add_arg("--rateest-lt", dest="rateest-lt", action="append")
    add_arg("--rateest-gt", dest="rateest-gt", action="append")
    add_arg("--rateest-eq", dest="rateest-eq", action="append")
    add_arg("--rateest-name", dest="rateest-name", action="append")
    add_arg("--rateest-interval", dest="rateest-interval", action="append")
    add_arg("--rateest-ewma", dest="rateest-ewma", action="append")
    ## realm
    add_arg("--realm", dest="realm", action="append")
    ## recent
    add_arg("--name", dest="name", action="append")
    add_arg("--set", dest="set", action="append")
    add_arg("--rsource", dest="rsource", action="append")
    add_arg("--rdest", dest="rdest", action="append")
    add_arg("--mask", dest="mask", action="append")
    add_arg("--rcheck", dest="rcheck", action="append")
    add_arg("--update", dest="update", action="append")
    add_arg("--remove", dest="remove", action="append")
    add_arg("--seconds", dest="seconds", action="append")
    add_arg("--reap", dest="reap", action="append")
    add_arg("--hitcount", dest="hitcount", action="append")
    add_arg("--rttl", dest="rttl", action="append")
    ## rpfilter
    add_arg("--loose", dest="loose", action="append")
    add_arg("--validmark", dest="validmark", action="append")
    add_arg("--accept-local", dest="accept-local", action="append")
    add_arg("--invert", dest="invert", action="append")
    ## rt
    add_arg("--rt-type", dest="rt-type", action="append")
    add_arg("--rt-segsleft", dest="rt-segsleft", action="append")
    add_arg("--rt-len", dest="rt-len", action="append")
    add_arg("--rt-0-res", dest="rt-0-res", action="append")
    add_arg("--rt-0-addrs", dest="rt-0-addrs", action="append")
    add_arg("--rt-0-not-strict", dest="rt-0-not-strict", action="append")
    ## sctp
    add_arg("--chunk-types", dest="chunk-types", action="append")
    ## set
    add_arg("--match-set", dest="match-set", action="append")
    add_arg("--return-nomatch", dest="return-nomatch", action="append")
    add_arg("--update-counters", dest="update-counters", action="append")
    add_arg("--update-subcounters", dest="update-subcounters", action="append")
    add_arg("--packets-eq", dest="packets-eq", action="append")
    add_arg("--packets-lt", dest="packets-lt", action="append")
    add_arg("--packets-gt", dest="packets-gt", action="append")
    add_arg("--bytes-eq", dest="bytes-eq", action="append")
    add_arg("--bytes-lt", dest="bytes-lt", action="append")
    add_arg("--bytes-gt", dest="bytes-gt", action="append")
    ## socket
    add_arg("--transparent", dest="transparent", action="append")
    add_arg("--nowildcard", dest="nowildcard", action="append")
    ## state
    add_arg("--state", dest="state", action="append")
    ## statistic
    add_arg("--probability", dest="probability", action="append")
    add_arg("--every", dest="every", action="append")
    add_arg("--packet", dest="packet", action="append")
    ## string
    add_arg("--algo", dest="algo", action="append")
    add_arg("--from", dest="from", action="append")
    add_arg("--to", dest="to", action="append")
    add_arg("--string", dest="string", action="append")
    add_arg("--hex-string", dest="hex-string", action="append")
    ## tcp
    add_arg("--tcp-flags", dest="tcp-flags", action="append")
    add_arg("--syn", dest="syn", action="append")
    add_arg("--tcp-option", dest="tcp-option", action="append")
    ## tcpmss
    add_arg("--mss", dest="mss", action="append")
    ## time
    add_arg("--datestart", dest="datestart", action="append")
    add_arg("--datestop", dest="datestop", action="append")
    add_arg("--timestart", dest="timestart", action="append")
    add_arg("--timestop", dest="timestop", action="append")
    add_arg("--monthdays", dest="monthdays", action="append")
    add_arg("--weekdays", dest="weekdays", action="append")
    add_arg("--contiguous", dest="contiguous", action="append")
    add_arg("--kerneltz", dest="kerneltz", action="append")
    add_arg("--utc", dest="utc", action="append")
    add_arg("--localtz", dest="localtz", action="append")
    ## tos
    add_arg("--tos", dest="tos", action="append")
    ## ttl
    add_arg("--ttl-eq", dest="ttl-eq", action="append")
    add_arg("--ttl-gt", dest="ttl-gt", action="append")
    add_arg("--ttl-lt", dest="ttl-lt", action="append")
    ## u32
    add_arg("--u32", dest="u32", action="append")

    # Xtables-addons matches
    ## condition
    add_arg("--condition", dest="condition", action="append")
    ## dhcpmac
    add_arg("--mac", dest="mac", action="append")
    ## fuzzy
    add_arg("--lower-limit", dest="lower-limit", action="append")
    add_arg("--upper-limit", dest="upper-limit", action="append")
    ## geoip
    add_arg("--src-cc", "--source-country", dest="source-country", action="append")
    add_arg(
        "--dst-cc", "--destination-country", dest="destination-country", action="append"
    )
    ## gradm
    add_arg("--enabled", dest="enabled", action="append")
    add_arg("--disabled", dest="disabled", action="append")
    ## iface
    add_arg("--iface", dest="iface", action="append")
    add_arg("--dev-in", dest="dev-in", action="append")
    add_arg("--dev-out", dest="dev-out", action="append")
    add_arg("--up", dest="up", action="append")
    add_arg("--down", dest="down", action="append")
    add_arg("--broadcast", dest="broadcast", action="append")
    add_arg("--loopback", dest="loopback", action="append")
    add_arg("--pointtopoint", dest="pointtopoint", action="append")
    add_arg("--running", dest="running", action="append")
    add_arg("--noarp", dest="noarp", action="append")
    add_arg("--arp", dest="arp", action="append")
    add_arg("--promisc", dest="promisc", action="append")
    add_arg("--multicast", dest="multicast", action="append")
    add_arg("--dynamic", dest="dynamic", action="append")
    add_arg("--lower-up", dest="lower-up", action="append")
    add_arg("--dormant", dest="dormant", action="append")
    ## ipp2p
    add_arg("--edk", dest="edk", action="append")
    add_arg("--kazaa", dest="kazaa", action="append")
    add_arg("--gnu", dest="gnu", action="append")
    add_arg("--dc", dest="dc", action="append")
    add_arg("--bit", dest="bit", action="append")
    add_arg("--apple", dest="apple", action="append")
    add_arg("--soul", dest="soul", action="append")
    add_arg("--winmx", dest="winmx", action="append")
    add_arg("--ares", dest="ares", action="append")
    add_arg("--debug", dest="debug", action="append")
    ## ipv4options
    add_arg("--flags", dest="flags", action="append")
    add_arg("--any", dest="any", action="append")
    ## length2
    add_arg("--layer3", dest="layer3", action="append")
    add_arg("--layer4", dest="layer4", action="append")
    add_arg("--layer5", dest="layer5", action="append")
    ## lscan
    add_arg("--stealth", dest="stealth", action="append")
    add_arg("--synscan", dest="synscan", action="append")
    add_arg("--cnscan", dest="cnscan", action="append")
    add_arg("--grscan", dest="grscan", action="append")
    ## psd
    add_arg("--psd-weight-threshold", dest="psd-weight-threshold", action="append")
    add_arg("--psd-delay-threshold", dest="psd-delay-threshold", action="append")
    add_arg("--psd-lo-ports-weight", dest="psd-lo-ports-weight", action="append")
    add_arg("--psd-hi-ports-weight", dest="psd-hi-ports-weight", action="append")
    ## quota2
    add_arg("--grow", dest="grow", action="append")
    add_arg("--no-change", dest="no-change", action="append")
    add_arg("--packets", dest="packets", action="append")
    ## pknock
    add_arg("--knockports", dest="knockports", action="append")
    add_arg("--time", dest="time", action="append")
    add_arg("--autoclose", dest="autoclose", action="append")
    add_arg("--checkip", dest="checkip", action="append")

    # TARGET EXTENSIONS
    ## AUDIT
    add_arg("--type", dest="type", action="append")
    ## CHECKSUM
    add_arg("--checksum-fill", dest="checksum-fill", action="append")
    ## CLASSIFY
    add_arg("--set-class", dest="set-class", action="append")
    ## CLUSTERIP
    add_arg("--new", dest="new", action="append")
    add_arg("--hashmode", dest="hashmode", action="append")
    add_arg("--clustermac", dest="clustermac", action="append")
    add_arg("--total-nodes", dest="total-nodes", action="append")
    add_arg("--local-node", dest="local-node", action="append")
    add_arg("--hash-init", dest="hash-init", action="append")
    ## CONNMARK
    add_arg("--set-xmark", dest="set-xmark", action="append")
    add_arg("--save-mark", dest="save-mark", action="append")
    add_arg("--restore-mark", dest="restore-mark", action="append")
    add_arg("--and-mark", dest="and-mark", action="append")
    add_arg("--or-mark", dest="or-mark", action="append")
    add_arg("--xor-mark", dest="xor-mark", action="append")
    add_arg("--set-mark", dest="set-mark", action="append")
    add_arg("--nfmask", dest="nfmask", action="append")
    add_arg("--ctmask", dest="ctmask", action="append")
    ## CONNSECMARK
    add_arg("--save", dest="save", action="append")
    add_arg("--restore", dest="restore", action="append")
    ## CT
    add_arg("--notrack", dest="notrack", action="append")
    add_arg("--ctevents", dest="ctevents", action="append")
    add_arg("--expevents", dest="expevents", action="append")
    add_arg("--zone", dest="zone", action="append")
    add_arg("--timeout", dest="timeout", action="append")
    ## DNAT
    add_arg("--to-destination", dest="to-destination", action="append")
    add_arg("--random", dest="random", action="append")
    add_arg("--persistent", dest="persistent", action="append")
    ## DNPT
    add_arg("--src-pfx", dest="src-pfx", action="append")
    add_arg("--dst-pfx", dest="dst-pfx", action="append")
    ## DSCP
    add_arg("--set-dscp", dest="set-dscp", action="append")
    add_arg("--set-dscp-class", dest="set-dscp-class", action="append")
    ## ECN
    add_arg("--ecn-tcp-remove", dest="ecn-tcp-remove", action="append")
    ## HL
    add_arg("--hl-set", dest="hl-set", action="append")
    add_arg("--hl-dec", dest="hl-dec", action="append")
    add_arg("--hl-inc", dest="hl-inc", action="append")
    ## HMARK
    add_arg("--hmark-tuple", dest="hmark-tuple", action="append")
    add_arg("--hmark-mod", dest="hmark-mod", action="append")
    add_arg("--hmark-offset", dest="hmark-offset", action="append")
    add_arg("--hmark-src-prefix", dest="hmark-src-prefix", action="append")
    add_arg("--hmark-dst-prefix", dest="hmark-dst-prefix", action="append")
    add_arg("--hmark-sport-mask", dest="hmark-sport-mask", action="append")
    add_arg("--hmark-dport-mask", dest="hmark-dport-mask", action="append")
    add_arg("--hmark-spi-mask", dest="hmark-spi-mask", action="append")
    add_arg("--hmark-proto-mask", dest="hmark-proto-mask", action="append")
    add_arg("--hmark-rnd", dest="hmark-rnd", action="append")
    ## LED
    add_arg("--led-trigger-id", dest="led-trigger-id", action="append")
    add_arg("--led-delay", dest="led-delay", action="append")
    add_arg("--led-always-blink", dest="led-always-blink", action="append")
    ## LOG
    add_arg("--log-level", dest="log-level", action="append")
    add_arg("--log-prefix", dest="log-prefix", action="append")
    add_arg("--log-tcp-sequence", dest="log-tcp-sequence", action="append")
    add_arg("--log-tcp-options", dest="log-tcp-options", action="append")
    add_arg("--log-ip-options", dest="log-ip-options", action="append")
    add_arg("--log-uid", dest="log-uid", action="append")
    ## MASQUERADE
    add_arg("--to-ports", dest="to-ports", action="append")
    ## NFLOG
    add_arg("--nflog-group", dest="nflog-group", action="append")
    add_arg("--nflog-prefix", dest="nflog-prefix", action="append")
    add_arg("--nflog-range", dest="nflog-range", action="append")
    add_arg("--nflog-threshold", dest="nflog-threshold", action="append")
    ## NFQUEUE
    add_arg("--queue-num", dest="queue-num", action="append")
    add_arg("--queue-balance", dest="queue-balance", action="append")
    add_arg("--queue-bypass", dest="queue-bypass", action="append")
    add_arg("--queue-cpu-fanout", dest="queue-cpu-fanout", action="append")
    ## RATEEST
    add_arg("--rateest-ewmalog", dest="rateest-ewmalog", action="append")
    ## REJECT
    add_arg("--reject-with", dest="reject-with", action="append")
    ## SAME
    add_arg("--nodst", dest="nodst", action="append")
    ## SECMARK
    add_arg("--selctx", dest="selctx", action="append")
    ## SET
    add_arg("--add-set", dest="add-set", action="append")
    add_arg("--del-set", dest="del-set", action="append")
    add_arg("--exist", dest="exist", action="append")
    ## SNAT
    add_arg("--to-source", dest="to-source", action="append")
    ## TCPMSS
    add_arg("--set-mss", dest="set-mss", action="append")
    add_arg("--clamp-mss-to-pmtu", dest="clamp-mss-to-pmtu", action="append")
    ## TCPOPTSTRIP
    add_arg("--strip-options", dest="strip-options", action="append")
    ## TEE
    add_arg("--gateway", dest="gateway", action="append")
    ## TOS
    add_arg("--set-tos", dest="set-tos", action="append")
    add_arg("--and-tos", dest="and-tos", action="append")
    add_arg("--or-tos", dest="or-tos", action="append")
    add_arg("--xor-tos", dest="xor-tos", action="append")
    ## TPROXY
    add_arg("--on-port", dest="on-port", action="append")
    add_arg("--on-ip", dest="on-ip", action="append")
    add_arg("--tproxy-mark", dest="tproxy-mark", action="append")
    ## TTL
    add_arg("--ttl-set", dest="ttl-set", action="append")
    add_arg("--ttl-dec", dest="ttl-dec", action="append")
    add_arg("--ttl-inc", dest="ttl-inc", action="append")
    ## ULOG
    add_arg("--ulog-nlgroup", dest="ulog-nlgroup", action="append")
    add_arg("--ulog-prefix", dest="ulog-prefix", action="append")
    add_arg("--ulog-cprange", dest="ulog-cprange", action="append")
    add_arg("--ulog-qthreshold", dest="ulog-qthreshold", action="append")

    # Xtables-addons targets
    ## ACCOUNT
    add_arg("--addr", dest="addr", action="append")
    add_arg("--tname", dest="tname", action="append")
    ## CHAOS
    add_arg("--delude", dest="delude", action="append")
    add_arg("--tarpit", dest="tarpit", action="append")
    ## DHCPMAC
    add_arg("--set-mac", dest="set-mac", action="append")
    ## DNETMAP
    add_arg("--prefix", dest="prefix", action="append")
    add_arg("--reuse", dest="reuse", action="append")
    add_arg("--static", dest="static", action="append")
    ## IPMARK
    add_arg("--and-mask", dest="and-mask", action="append")
    add_arg("--or-mask", dest="or-mask", action="append")
    add_arg("--shift", dest="shift", action="append")
    ## TARPIT
    add_arg("--honeypot", dest="honeypot", action="append")
    add_arg("--reset", dest="reset", action="append")

    return parser



"""
Read lm-sensors

.. versionadded:: 2014.1.3
"""

import logging

import salt.utils.path

log = logging.getLogger(__name__)


def __virtual__():
    if salt.utils.path.which("sensors"):
        return True
    return (False, "sensors does not exist in the path")


def sense(chip, fahrenheit=False):
    """
    Gather lm-sensors data from a given chip

    To determine the chip to query, use the 'sensors' command
    and see the leading line in the block.

    Example:

    /usr/bin/sensors

    coretemp-isa-0000
    Adapter: ISA adapter
    Physical id 0:  +56.0°C  (high = +87.0°C, crit = +105.0°C)
    Core 0:         +52.0°C  (high = +87.0°C, crit = +105.0°C)
    Core 1:         +50.0°C  (high = +87.0°C, crit = +105.0°C)
    Core 2:         +56.0°C  (high = +87.0°C, crit = +105.0°C)
    Core 3:         +53.0°C  (high = +87.0°C, crit = +105.0°C)

    Given the above, the chip is 'coretemp-isa-0000'.
    """
    extra_args = ""
    if fahrenheit is True:
        extra_args = "-f"
    sensors = __salt__["cmd.run"](
        "/usr/bin/sensors {} {}".format(chip, extra_args), python_shell=False
    ).splitlines()
    ret = {}
    for sensor in sensors:
        sensor_list = sensor.split(":")
        if len(sensor_list) >= 2:
            ret[sensor_list[0]] = sensor_list[1].lstrip()
    return ret



import subprocess

import salt.modules.openscap as openscap
from tests.support.mock import MagicMock, Mock, patch
from tests.support.unit import TestCase


class OpenscapTestCase(TestCase):

    random_temp_dir = "/tmp/unique-name"
    policy_file = "/usr/share/openscap/policy-file-xccdf.xml"

    def setUp(self):
        import salt.modules.openscap

        salt.modules.openscap.__salt__ = MagicMock()
        patchers = [
            patch("salt.modules.openscap.__salt__", MagicMock()),
            patch("salt.modules.openscap.shutil.rmtree", Mock()),
            patch(
                "salt.modules.openscap.tempfile.mkdtemp",
                Mock(return_value=self.random_temp_dir),
            ),
        ]
        for patcher in patchers:
            self.apply_patch(patcher)

    def apply_patch(self, patcher):
        patcher.start()
        self.addCleanup(patcher.stop)

    def test_openscap_xccdf_eval_success(self):
        with patch(
            "salt.modules.openscap.Popen",
            MagicMock(
                return_value=Mock(
                    **{"returncode": 0, "communicate.return_value": ("", "")}
                )
            ),
        ):
            response = openscap.xccdf(
                "eval --profile Default {}".format(self.policy_file)
            )

            self.assertEqual(openscap.tempfile.mkdtemp.call_count, 1)
            expected_cmd = [
                "oscap",
                "xccdf",
                "eval",
                "--oval-results",
                "--results",
                "results.xml",
                "--report",
                "report.html",
                "--profile",
                "Default",
                self.policy_file,
            ]
            openscap.Popen.assert_called_once_with(
                expected_cmd,
                cwd=openscap.tempfile.mkdtemp.return_value,
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE,
            )
            openscap.__salt__["cp.push_dir"].assert_called_once_with(
                self.random_temp_dir
            )
            self.assertEqual(openscap.shutil.rmtree.call_count, 1)
            self.assertEqual(
                response,
                {
                    "upload_dir": self.random_temp_dir,
                    "error": "",
                    "success": True,
                    "returncode": 0,
                },
            )

    def test_openscap_xccdf_eval_success_with_failing_rules(self):
        with patch(
            "salt.modules.openscap.Popen",
            MagicMock(
                return_value=Mock(
                    **{"returncode": 2, "communicate.return_value": ("", "some error")}
                )
            ),
        ):
            response = openscap.xccdf(
                "eval --profile Default {}".format(self.policy_file)
            )

            self.assertEqual(openscap.tempfile.mkdtemp.call_count, 1)
            expected_cmd = [
                "oscap",
                "xccdf",
                "eval",
                "--oval-results",
                "--results",
                "results.xml",
                "--report",
                "report.html",
                "--profile",
                "Default",
                self.policy_file,
            ]
            openscap.Popen.assert_called_once_with(
                expected_cmd,
                cwd=openscap.tempfile.mkdtemp.return_value,
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE,
            )
            openscap.__salt__["cp.push_dir"].assert_called_once_with(
                self.random_temp_dir
            )
            self.assertEqual(openscap.shutil.rmtree.call_count, 1)
            self.assertEqual(
                response,
                {
                    "upload_dir": self.random_temp_dir,
                    "error": "some error",
                    "success": True,
                    "returncode": 2,
                },
            )

    def test_openscap_xccdf_eval_fail_no_profile(self):
        response = openscap.xccdf("eval --param Default /unknown/param")
        error = "the following arguments are required: --profile"
        self.assertEqual(
            response,
            {"error": error, "upload_dir": None, "success": False, "returncode": None},
        )

    def test_openscap_xccdf_eval_success_ignore_unknown_params(self):
        with patch(
            "salt.modules.openscap.Popen",
            MagicMock(
                return_value=Mock(
                    **{"returncode": 2, "communicate.return_value": ("", "some error")}
                )
            ),
        ):
            response = openscap.xccdf(
                "eval --profile Default --param Default /policy/file"
            )
            self.assertEqual(
                response,
                {
                    "upload_dir": self.random_temp_dir,
                    "error": "some error",
                    "success": True,
                    "returncode": 2,
                },
            )
            expected_cmd = [
                "oscap",
                "xccdf",
                "eval",
                "--oval-results",
                "--results",
                "results.xml",
                "--report",
                "report.html",
                "--profile",
                "Default",
                "/policy/file",
            ]
            openscap.Popen.assert_called_once_with(
                expected_cmd,
                cwd=openscap.tempfile.mkdtemp.return_value,
                stderr=subprocess.PIPE,
                stdout=subprocess.PIPE,
            )

    def test_openscap_xccdf_eval_evaluation_error(self):
        with patch(
            "salt.modules.openscap.Popen",
            MagicMock(
                return_value=Mock(
                    **{
                        "returncode": 1,
                        "communicate.return_value": ("", "evaluation error"),
                    }
                )
            ),
        ):
            response = openscap.xccdf(
                "eval --profile Default {}".format(self.policy_file)
            )

            self.assertEqual(
                response,
                {
                    "upload_dir": None,
                    "error": "evaluation error",
                    "success": False,
                    "returncode": 1,
                },
            )

    def test_openscap_xccdf_eval_fail_not_implemented_action(self):
        response = openscap.xccdf("info {}".format(self.policy_file))
        mock_err = "argument action: invalid choice: 'info' (choose from 'eval')"

        self.assertEqual(
            response,
            {
                "upload_dir": None,
                "error": mock_err,
                "success": False,
                "returncode": None,
            },
        )



"""
    :codeauthor: Mike Place <mp@saltstack.com>
"""


import errno
import logging
import os
import threading

import pytest
import salt.config
import salt.exceptions
import salt.ext.tornado.gen
import salt.ext.tornado.ioloop
import salt.ext.tornado.testing
import salt.transport.client
import salt.transport.ipc
import salt.transport.server
import salt.utils.platform
from salt.ext.tornado.iostream import StreamClosedError
from tests.support.runtests import RUNTIME_VARS
from tests.support.unit import skipIf

pytestmark = [
    pytest.mark.skip_on_darwin,
    pytest.mark.skip_on_freebsd,
    pytest.mark.skip_on_windows,
]

log = logging.getLogger(__name__)


@skipIf(salt.utils.platform.is_windows(), "Windows does not support Posix IPC")
class IPCMessagePubSubCase(salt.ext.tornado.testing.AsyncTestCase):
    """
    Test all of the clear msg stuff
    """

    def setUp(self):
        super().setUp()
        self.opts = {"ipc_write_buffer": 0}
        if not os.path.exists(RUNTIME_VARS.TMP):
            os.mkdir(RUNTIME_VARS.TMP)
        self.socket_path = os.path.join(RUNTIME_VARS.TMP, "ipc_test.ipc")
        self.pub_channel = self._get_pub_channel()
        self.sub_channel = self._get_sub_channel()

    def _get_pub_channel(self):
        pub_channel = salt.transport.ipc.IPCMessagePublisher(
            self.opts,
            self.socket_path,
        )
        pub_channel.start()
        return pub_channel

    def _get_sub_channel(self):
        sub_channel = salt.transport.ipc.IPCMessageSubscriber(
            socket_path=self.socket_path,
            io_loop=self.io_loop,
        )
        sub_channel.connect(callback=self.stop)
        self.wait()
        return sub_channel

    def tearDown(self):
        super().tearDown()
        try:
            self.pub_channel.close()
        except OSError as exc:
            if exc.errno != errno.EBADF:
                # If its not a bad file descriptor error, raise
                raise
        try:
            self.sub_channel.close()
        except OSError as exc:
            if exc.errno != errno.EBADF:
                # If its not a bad file descriptor error, raise
                raise
        os.unlink(self.socket_path)
        del self.pub_channel
        del self.sub_channel

    def test_multi_client_reading(self):
        # To be completely fair let's create 2 clients.
        client1 = self.sub_channel
        client2 = self._get_sub_channel()
        call_cnt = []

        # Create a watchdog to be safe from hanging in sync loops (what old code did)
        evt = threading.Event()

        def close_server():
            if evt.wait(1):
                return
            client2.close()
            self.stop()

        watchdog = threading.Thread(target=close_server)
        watchdog.start()

        # Runs in ioloop thread so we're safe from race conditions here
        def handler(raw):
            call_cnt.append(raw)
            if len(call_cnt) >= 2:
                evt.set()
                self.stop()

        # Now let both waiting data at once
        client1.read_async(handler)
        client2.read_async(handler)
        self.pub_channel.publish("TEST")
        self.wait()
        self.assertEqual(len(call_cnt), 2)
        self.assertEqual(call_cnt[0], "TEST")
        self.assertEqual(call_cnt[1], "TEST")

    def test_sync_reading(self):
        # To be completely fair let's create 2 clients.
        client1 = self.sub_channel
        client2 = self._get_sub_channel()
        call_cnt = []

        # Now let both waiting data at once
        self.pub_channel.publish("TEST")
        ret1 = client1.read_sync()
        ret2 = client2.read_sync()
        self.assertEqual(ret1, "TEST")
        self.assertEqual(ret2, "TEST")

    @salt.ext.tornado.testing.gen_test
    def test_async_reading_streamclosederror(self):
        client1 = self.sub_channel
        call_cnt = []

        # Create a watchdog to be safe from hanging in sync loops (what old code did)
        evt = threading.Event()

        def close_server():
            if evt.wait(0.001):
                return
            client1.close()
            self.stop()

        watchdog = threading.Thread(target=close_server)
        watchdog.start()

        # Runs in ioloop thread so we're safe from race conditions here
        def handler(raw):
            pass

        try:
            ret1 = yield client1.read_async(handler)
            self.wait()
        except StreamClosedError as ex:
            assert False, "StreamClosedError was raised inside the Future"



"""
    :codeauthor: Rahul Handay <rahulha@saltstack.com>
"""

import os

import pytest
import salt.config
import salt.states.winrepo as winrepo
import salt.utils.path
from salt.syspaths import BASE_FILE_ROOTS_DIR
from tests.support.mock import MagicMock, patch


class MockRunnerClient:
    """
    Mock RunnerClient class
    """

    def __init__(self):
        pass

    class RunnerClient:
        """
        Mock RunnerClient class
        """

        def __init__(self, master_config):
            """
            init method
            """

        @staticmethod
        def cmd(arg1, arg2):
            """
            Mock cmd method
            """
            # TODO: Figure out how to have this return an empty dict or a dict
            # TODO: with expected data
            return []


@pytest.fixture
def configure_loader_modules():
    patcher = patch("salt.states.winrepo.salt.runner", MockRunnerClient)
    patcher.start()
    return {winrepo: {}}


def test_genrepo():
    """
    Test to refresh the winrepo.p file of the repository
    """
    expected = {"name": "salt", "changes": {}, "result": False, "comment": ""}
    mock_config = MagicMock(
        return_value={"winrepo_dir": "salt", "winrepo_cachefile": "abc"}
    )
    mock_stat = MagicMock(return_value=[0, 1, 2, 3, 4, 5, 6, 7, 8])
    mock_empty_list = MagicMock(return_value=[])
    with patch.object(salt.config, "master_config", mock_config), patch.object(
        os, "stat", mock_stat
    ), patch.object(salt.utils.path, "os_walk", mock_empty_list), patch.dict(
        winrepo.__opts__, {"test": True}
    ):
        # With test=True
        expected.update({"comment": "", "result": None})
        assert winrepo.genrepo("salt") == expected

        with patch.dict(winrepo.__opts__, {"test": False}):
            # With test=False
            expected.update({"result": True})
            assert winrepo.genrepo("salt") == expected

            # Now with no changes, existing winrepo.p
            expected.update({"changes": {"winrepo": []}})
            assert winrepo.genrepo("salt", True) == expected


def test_genrepo_no_dir():
    """
    Test genrepo when the dir does not exist
    """
    expected = {
        "name": "salt",
        "changes": {},
        "result": False,
        "comment": "{} is missing".format(
            os.sep.join([BASE_FILE_ROOTS_DIR, "win", "repo"])
        ),
    }
    with patch.dict(winrepo.__opts__, {"test": False}), patch(
        "os.path.exists", MagicMock(return_value=False)
    ):
        ret = winrepo.genrepo("salt")
        assert ret == expected


def test_genrepo_no_dir_force():
    """
    Test genrepo when the dir does not exist and force=True
    """
    expected = {
        "name": "salt",
        "changes": {"winrepo": []},
        "result": True,
        "comment": "",
    }
    with patch.dict(winrepo.__opts__, {"test": False}), patch(
        "os.path.exists", MagicMock(return_value=False)
    ):
        ret = winrepo.genrepo("salt", force=True)
        assert ret == expected



"""
    :codeauthor: :email:`Alexandru Bleotu <alexandru.bleotu@morganstanley.com>`

    Tests for esxdatacenter proxy
"""


import salt.exceptions
import salt.proxy.esxdatacenter as esxdatacenter
from salt.config.schemas.esxdatacenter import EsxdatacenterProxySchema
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf

try:
    import jsonschema

    HAS_JSONSCHEMA = True
except ImportError:
    HAS_JSONSCHEMA = False


@skipIf(not HAS_JSONSCHEMA, "jsonschema is required")
class InitTestCase(TestCase, LoaderModuleMockMixin):
    """Tests for salt.proxy.esxdatacenter.init"""

    def setup_loader_modules(self):
        return {esxdatacenter: {"DETAILS": {}, "__pillar__": {}}}

    def setUp(self):
        self.opts_userpass = {
            "proxy": {
                "proxytype": "esxdatacenter",
                "vcenter": "fake_vcenter",
                "datacenter": "fake_dc",
                "mechanism": "userpass",
                "username": "fake_username",
                "passwords": ["fake_password"],
                "protocol": "fake_protocol",
                "port": 100,
            }
        }
        self.opts_sspi = {
            "proxy": {
                "proxytype": "esxdatacenter",
                "vcenter": "fake_vcenter",
                "datacenter": "fake_dc",
                "mechanism": "sspi",
                "domain": "fake_domain",
                "principal": "fake_principal",
                "protocol": "fake_protocol",
                "port": 100,
            }
        }
        patches = (
            (
                "salt.proxy.esxdatacenter.merge",
                MagicMock(return_value=self.opts_sspi["proxy"]),
            ),
        )
        for mod, mock in patches:
            patcher = patch(mod, mock)
            patcher.start()
            self.addCleanup(patcher.stop)

    def test_merge(self):
        mock_pillar_proxy = MagicMock()
        mock_opts_proxy = MagicMock()
        mock_merge = MagicMock(return_value=self.opts_sspi["proxy"])
        with patch.dict(esxdatacenter.__pillar__, {"proxy": mock_pillar_proxy}):
            with patch("salt.proxy.esxdatacenter.merge", mock_merge):
                esxdatacenter.init(opts={"proxy": mock_opts_proxy})
        mock_merge.assert_called_once_with(mock_opts_proxy, mock_pillar_proxy)

    def test_esxdatacenter_schema(self):
        mock_json_validate = MagicMock()
        serialized_schema = EsxdatacenterProxySchema().serialize()
        with patch("salt.proxy.esxdatacenter.jsonschema.validate", mock_json_validate):
            esxdatacenter.init(self.opts_sspi)
        mock_json_validate.assert_called_once_with(
            self.opts_sspi["proxy"], serialized_schema
        )

    def test_invalid_proxy_input_error(self):
        with patch(
            "salt.proxy.esxdatacenter.jsonschema.validate",
            MagicMock(
                side_effect=jsonschema.exceptions.ValidationError("Validation Error")
            ),
        ):
            with self.assertRaises(salt.exceptions.InvalidConfigError) as excinfo:
                esxdatacenter.init(self.opts_userpass)
        self.assertEqual(excinfo.exception.strerror, "Validation Error")

    def test_no_username(self):
        opts = self.opts_userpass.copy()
        del opts["proxy"]["username"]
        with patch(
            "salt.proxy.esxdatacenter.merge", MagicMock(return_value=opts["proxy"])
        ):
            with self.assertRaises(salt.exceptions.InvalidConfigError) as excinfo:
                esxdatacenter.init(opts)
        self.assertEqual(
            excinfo.exception.strerror,
            "Mechanism is set to 'userpass', but no "
            "'username' key found in proxy config.",
        )

    def test_no_passwords(self):
        opts = self.opts_userpass.copy()
        del opts["proxy"]["passwords"]
        with patch(
            "salt.proxy.esxdatacenter.merge", MagicMock(return_value=opts["proxy"])
        ):
            with self.assertRaises(salt.exceptions.InvalidConfigError) as excinfo:
                esxdatacenter.init(opts)
        self.assertEqual(
            excinfo.exception.strerror,
            "Mechanism is set to 'userpass', but no "
            "'passwords' key found in proxy config.",
        )

    def test_no_domain(self):
        opts = self.opts_sspi.copy()
        del opts["proxy"]["domain"]
        with patch(
            "salt.proxy.esxdatacenter.merge", MagicMock(return_value=opts["proxy"])
        ):
            with self.assertRaises(salt.exceptions.InvalidConfigError) as excinfo:
                esxdatacenter.init(opts)
        self.assertEqual(
            excinfo.exception.strerror,
            "Mechanism is set to 'sspi', but no 'domain' key found in proxy config.",
        )

    def test_no_principal(self):
        opts = self.opts_sspi.copy()
        del opts["proxy"]["principal"]
        with patch(
            "salt.proxy.esxdatacenter.merge", MagicMock(return_value=opts["proxy"])
        ):
            with self.assertRaises(salt.exceptions.InvalidConfigError) as excinfo:
                esxdatacenter.init(opts)
        self.assertEqual(
            excinfo.exception.strerror,
            "Mechanism is set to 'sspi', but no 'principal' key found in proxy config.",
        )

    def test_find_credentials(self):
        mock_find_credentials = MagicMock(
            return_value=("fake_username", "fake_password")
        )
        with patch(
            "salt.proxy.esxdatacenter.merge",
            MagicMock(return_value=self.opts_userpass["proxy"]),
        ):
            with patch(
                "salt.proxy.esxdatacenter.find_credentials", mock_find_credentials
            ):
                esxdatacenter.init(self.opts_userpass)
        mock_find_credentials.assert_called_once_with()

    def test_details_userpass(self):
        mock_find_credentials = MagicMock(
            return_value=("fake_username", "fake_password")
        )
        with patch(
            "salt.proxy.esxdatacenter.merge",
            MagicMock(return_value=self.opts_userpass["proxy"]),
        ):
            with patch(
                "salt.proxy.esxdatacenter.find_credentials", mock_find_credentials
            ):
                esxdatacenter.init(self.opts_userpass)
        self.assertDictEqual(
            esxdatacenter.DETAILS,
            {
                "vcenter": "fake_vcenter",
                "datacenter": "fake_dc",
                "mechanism": "userpass",
                "username": "fake_username",
                "password": "fake_password",
                "passwords": ["fake_password"],
                "protocol": "fake_protocol",
                "port": 100,
            },
        )

    def test_details_sspi(self):
        esxdatacenter.init(self.opts_sspi)
        self.assertDictEqual(
            esxdatacenter.DETAILS,
            {
                "vcenter": "fake_vcenter",
                "datacenter": "fake_dc",
                "mechanism": "sspi",
                "domain": "fake_domain",
                "principal": "fake_principal",
                "protocol": "fake_protocol",
                "port": 100,
            },
        )



"""
    :codeauthor: Nicole Thomas <nicole@saltstack.com>
"""

import logging

import pytest  # pylint: disable=unused-import
import salt.state

log = logging.getLogger(__name__)


@pytest.fixture
def root_dir(tmp_path):
    return tmp_path / "root_dir"


@pytest.fixture
def state_tree_dir(root_dir):
    return root_dir / "state_tree"


@pytest.fixture
def cache_dir(root_dir):
    return root_dir / "cache_dir"


@pytest.fixture
def highstate(temp_salt_minion, temp_salt_master, root_dir, state_tree_dir, cache_dir):
    # for dpath in (root_dir, state_tree_dir, cache_dir):
    #    if not os.path.isdir(dpath):
    #        os.makedirs(dpath)

    opts = temp_salt_minion.config.copy()
    opts["root_dir"] = str(root_dir)
    opts["state_events"] = False
    opts["id"] = "match"
    opts["file_client"] = "local"
    opts["file_roots"] = dict(base=[str(state_tree_dir)])
    opts["cachedir"] = str(cache_dir)
    opts["test"] = False

    opts.update(
        {
            "transport": "zeromq",
            "auth_tries": 1,
            "auth_timeout": 5,
            "master_ip": "127.0.0.1",
            "master_port": temp_salt_master.config["ret_port"],
            "master_uri": "tcp://127.0.0.1:{}".format(
                temp_salt_master.config["ret_port"]
            ),
        }
    )

    _highstate = salt.state.HighState(opts)
    _highstate.push_active()

    yield _highstate


def test_top_matches_with_list(highstate):
    top = {"env": {"match": ["state1", "state2"], "nomatch": ["state3"]}}
    matches = highstate.top_matches(top)
    assert matches == {"env": ["state1", "state2"]}


def test_top_matches_with_string(highstate):
    top = {"env": {"match": "state1", "nomatch": "state2"}}
    matches = highstate.top_matches(top)
    assert matches == {"env": ["state1"]}


def test_matches_whitelist(highstate):
    matches = {"env": ["state1", "state2", "state3"]}
    matches = highstate.matches_whitelist(matches, ["state2"])
    assert matches == {"env": ["state2"]}


def test_matches_whitelist_with_string(highstate):
    matches = {"env": ["state1", "state2", "state3"]}
    matches = highstate.matches_whitelist(matches, "state2,state3")
    assert matches == {"env": ["state2", "state3"]}


def test_compile_state_usage(highstate, state_tree_dir):
    top = pytest.helpers.temp_file("top.sls", "base: {'*': [foo]}", str(state_tree_dir))
    used_state = pytest.helpers.temp_file(
        "foo.sls", "foo: test.nop", str(state_tree_dir)
    )
    unused_state = pytest.helpers.temp_file(
        "bar.sls", "bar: test.nop", str(state_tree_dir)
    )

    with top, used_state, unused_state:
        state_usage_dict = highstate.compile_state_usage()

        assert state_usage_dict["base"]["count_unused"] == 2
        assert state_usage_dict["base"]["count_used"] == 1
        assert state_usage_dict["base"]["count_all"] == 3
        assert state_usage_dict["base"]["used"] == ["foo"]
        assert state_usage_dict["base"]["unused"] == ["bar", "top"]


def test_find_sls_ids_with_exclude(highstate, state_tree_dir):
    """
    See https://github.com/saltstack/salt/issues/47182
    """
    top_sls = """
    base:
      '*':
        - issue-47182.stateA
        - issue-47182.stateB
        """

    slsfile1 = """
    slsfile1-nop:
        test.nop
        """

    slsfile2 = """
    slsfile2-nop:
      test.nop
        """

    stateB = """
    include:
      - issue-47182.slsfile1
      - issue-47182.slsfile2

    some-state:
      test.nop:
        - require:
          - sls: issue-47182.slsfile1
        - require_in:
          - sls: issue-47182.slsfile2
        """

    stateA_init = """
    include:
      - issue-47182.stateA.newer
        """

    stateA_newer = """
    exclude:
      - sls: issue-47182.stateA

    somestuff:
      cmd.run:
        - name: echo This supersedes the stuff previously done in issue-47182.stateA
        """

    sls_dir = str(state_tree_dir / "issue-47182")
    stateA_sls_dir = str(state_tree_dir / "issue-47182" / "stateA")

    with pytest.helpers.temp_file("top.sls", top_sls, str(state_tree_dir)):
        with pytest.helpers.temp_file(
            "slsfile1.sls", slsfile1, sls_dir
        ), pytest.helpers.temp_file(
            "slsfile1.sls", slsfile1, sls_dir
        ), pytest.helpers.temp_file(
            "stateB.sls", stateB, sls_dir
        ), pytest.helpers.temp_file(
            "init.sls", stateA_init, stateA_sls_dir
        ), pytest.helpers.temp_file(
            "newer.sls", stateA_newer, stateA_sls_dir
        ):
            # Manually compile the high data. We don't have to worry about all of
            # the normal error checking we do here since we know that all the SLS
            # files exist and there is no whitelist/blacklist being used.
            top = highstate.get_top()  # pylint: disable=assignment-from-none
            matches = highstate.top_matches(top)
            high, _ = highstate.render_highstate(matches)
            ret = salt.state.find_sls_ids("issue-47182.stateA.newer", high)
            assert ret == [("somestuff", "cmd")]



"""
Wrapper module for at(1)

Also, a 'tag' feature has been added to more
easily tag jobs.

:platform:      linux,openbsd,freebsd

.. versionchanged:: 2017.7.0
"""

import datetime
import re
import time

import salt.utils.data
import salt.utils.path
import salt.utils.platform

# pylint: enable=import-error,redefined-builtin
from salt.exceptions import CommandNotFoundError

# pylint: disable=import-error,redefined-builtin

# OS Families that should work (Ubuntu and Debian are the default)
# TODO: Refactor some of this module to remove the checks for binaries

# Tested on OpenBSD 5.0
BSD = ("OpenBSD", "FreeBSD")

__virtualname__ = "at"


def __virtual__():
    """
    Most everything has the ability to support at(1)
    """
    if salt.utils.platform.is_windows() or salt.utils.platform.is_sunos():
        return (False, "The at module could not be loaded: unsupported platform")
    if salt.utils.path.which("at") is None:
        return (False, "The at module could not be loaded: at command not found")
    return __virtualname__


def _cmd(binary, *args):
    """
    Wrapper to run at(1) or return None.
    """
    binary = salt.utils.path.which(binary)
    if not binary:
        raise CommandNotFoundError("{}: command not found".format(binary))
    cmd = [binary] + list(args)
    return __salt__["cmd.run_stdout"]([binary] + list(args), python_shell=False)


def atq(tag=None):
    """
    List all queued and running jobs or only those with
    an optional 'tag'.

    CLI Example:

    .. code-block:: bash

        salt '*' at.atq
        salt '*' at.atq [tag]
        salt '*' at.atq [job number]
    """
    jobs = []

    # Shim to produce output similar to what __virtual__() should do
    # but __salt__ isn't available in __virtual__()
    # Tested on CentOS 5.8
    if __grains__["os_family"] == "RedHat":
        output = _cmd("at", "-l")
    else:
        output = _cmd("atq")

    if output is None:
        return "'at.atq' is not available."

    # No jobs so return
    if output == "":
        return {"jobs": jobs}

    # Jobs created with at.at() will use the following
    # comment to denote a tagged job.
    job_kw_regex = re.compile(r"^### SALT: (\w+)")

    # Split each job into a dictionary and handle
    # pulling out tags or only listing jobs with a certain
    # tag
    for line in output.splitlines():
        job_tag = ""

        # Redhat/CentOS
        if __grains__["os_family"] == "RedHat":
            job, spec = line.split("\t")
            specs = spec.split()
        elif __grains__["os"] == "OpenBSD":
            if line.startswith(" Rank"):
                continue
            else:
                tmp = line.split()
                timestr = " ".join(tmp[1:5])
                job = tmp[6]
                specs = (
                    datetime.datetime(*(time.strptime(timestr, "%b %d, %Y %H:%M")[0:5]))
                    .isoformat()
                    .split("T")
                )
                specs.append(tmp[7])
                specs.append(tmp[5])
        elif __grains__["os"] == "FreeBSD":
            if line.startswith("Date"):
                continue
            else:
                tmp = line.split()
                timestr = " ".join(tmp[1:6])
                job = tmp[8]
                specs = (
                    datetime.datetime(
                        *(time.strptime(timestr, "%b %d %H:%M:%S %Z %Y")[0:5])
                    )
                    .isoformat()
                    .split("T")
                )
                specs.append(tmp[7])
                specs.append(tmp[6])

        else:
            job, spec = line.split("\t")
            tmp = spec.split()
            timestr = " ".join(tmp[0:5])
            specs = (
                datetime.datetime(*(time.strptime(timestr)[0:5])).isoformat().split("T")
            )
            specs.append(tmp[5])
            specs.append(tmp[6])

        # Search for any tags
        atc_out = _cmd("at", "-c", job)
        for line in atc_out.splitlines():
            tmp = job_kw_regex.match(line)
            if tmp:
                job_tag = tmp.groups()[0]

        if __grains__["os"] in BSD:
            job = str(job)
        else:
            job = int(job)

        # If a tag is supplied, only list jobs with that tag
        if tag:
            # TODO: Looks like there is a difference between salt and salt-call
            # If I don't wrap job in an int(), it fails on salt but works on
            # salt-call. With the int(), it fails with salt-call but not salt.
            if tag == job_tag or tag == job:
                jobs.append(
                    {
                        "job": job,
                        "date": specs[0],
                        "time": specs[1],
                        "queue": specs[2],
                        "user": specs[3],
                        "tag": job_tag,
                    }
                )
        else:
            jobs.append(
                {
                    "job": job,
                    "date": specs[0],
                    "time": specs[1],
                    "queue": specs[2],
                    "user": specs[3],
                    "tag": job_tag,
                }
            )

    return {"jobs": jobs}


def atrm(*args):
    """
    Remove jobs from the queue.

    CLI Example:

    .. code-block:: bash

        salt '*' at.atrm <jobid> <jobid> .. <jobid>
        salt '*' at.atrm all
        salt '*' at.atrm all [tag]
    """

    # Need to do this here also since we use atq()
    if not salt.utils.path.which("at"):
        return "'at.atrm' is not available."

    if not args:
        return {"jobs": {"removed": [], "tag": None}}

    # Convert all to strings
    args = salt.utils.data.stringify(args)

    if args[0] == "all":
        if len(args) > 1:
            opts = list(list(map(str, [j["job"] for j in atq(args[1])["jobs"]])))
            ret = {"jobs": {"removed": opts, "tag": args[1]}}
        else:
            opts = list(list(map(str, [j["job"] for j in atq()["jobs"]])))
            ret = {"jobs": {"removed": opts, "tag": None}}
    else:
        opts = list(
            list(
                map(
                    str,
                    [i["job"] for i in atq()["jobs"] if str(i["job"]) in args],
                )
            )
        )
        ret = {"jobs": {"removed": opts, "tag": None}}

    # Shim to produce output similar to what __virtual__() should do
    # but __salt__ isn't available in __virtual__()
    output = _cmd("at", "-d", " ".join(opts))
    if output is None:
        return "'at.atrm' is not available."

    return ret


def at(*args, **kwargs):  # pylint: disable=C0103
    """
    Add a job to the queue.

    The 'timespec' follows the format documented in the
    at(1) manpage.

    CLI Example:

    .. code-block:: bash

        salt '*' at.at <timespec> <cmd> [tag=<tag>] [runas=<user>]
        salt '*' at.at 12:05am '/sbin/reboot' tag=reboot
        salt '*' at.at '3:05am +3 days' 'bin/myscript' tag=nightly runas=jim
    """

    if len(args) < 2:
        return {"jobs": []}

    # Shim to produce output similar to what __virtual__() should do
    # but __salt__ isn't available in __virtual__()
    binary = salt.utils.path.which("at")
    if not binary:
        return "'at.at' is not available."

    if "tag" in kwargs:
        stdin = "### SALT: {}\n{}".format(kwargs["tag"], " ".join(args[1:]))
    else:
        stdin = " ".join(args[1:])
    cmd = [binary, args[0]]

    cmd_kwargs = {"stdin": stdin, "python_shell": False}
    if "runas" in kwargs:
        cmd_kwargs["runas"] = kwargs["runas"]
    output = __salt__["cmd.run"](cmd, **cmd_kwargs)

    if output is None:
        return "'at.at' is not available."

    if output.endswith("Garbled time"):
        return {"jobs": [], "error": "invalid timespec"}

    if output.startswith("warning: commands"):
        output = output.splitlines()[1]

    if output.startswith("commands will be executed"):
        output = output.splitlines()[1]

    output = output.split()[1]

    if __grains__["os"] in BSD:
        return atq(str(output))
    else:
        return atq(int(output))


def atc(jobid):
    """
    Print the at(1) script that will run for the passed job
    id. This is mostly for debugging so the output will
    just be text.

    CLI Example:

    .. code-block:: bash

        salt '*' at.atc <jobid>
    """
    # Shim to produce output similar to what __virtual__() should do
    # but __salt__ isn't available in __virtual__()
    output = _cmd("at", "-c", str(jobid))

    if output is None:
        return "'at.atc' is not available."
    elif output == "":
        return {"error": "invalid job id '{}'".format(jobid)}

    return output


def _atq(**kwargs):
    """
    Return match jobs list
    """

    jobs = []

    runas = kwargs.get("runas", None)
    tag = kwargs.get("tag", None)
    hour = kwargs.get("hour", None)
    minute = kwargs.get("minute", None)
    day = kwargs.get("day", None)
    month = kwargs.get("month", None)
    year = kwargs.get("year", None)
    if year and len(str(year)) == 2:
        year = "20{}".format(year)

    jobinfo = atq()["jobs"]
    if not jobinfo:
        return {"jobs": jobs}

    for job in jobinfo:

        if not runas:
            pass
        elif runas == job["user"]:
            pass
        else:
            continue

        if not tag:
            pass
        elif tag == job["tag"]:
            pass
        else:
            continue

        if not hour:
            pass
        elif "{:02d}".format(int(hour)) == job["time"].split(":")[0]:
            pass
        else:
            continue

        if not minute:
            pass
        elif "{:02d}".format(int(minute)) == job["time"].split(":")[1]:
            pass
        else:
            continue

        if not day:
            pass
        elif "{:02d}".format(int(day)) == job["date"].split("-")[2]:
            pass
        else:
            continue

        if not month:
            pass
        elif "{:02d}".format(int(month)) == job["date"].split("-")[1]:
            pass
        else:
            continue

        if not year:
            pass
        elif year == job["date"].split("-")[0]:
            pass
        else:
            continue

        jobs.append(job)

    if not jobs:
        note = "No match jobs or time format error"
        return {"jobs": jobs, "note": note}

    return {"jobs": jobs}


def jobcheck(**kwargs):
    """
    Check the job from queue.
    The kwargs dict include 'hour minute day month year tag runas'
    Other parameters will be ignored.

    CLI Example:

    .. code-block:: bash

        salt '*' at.jobcheck runas=jam day=13
        salt '*' at.jobcheck day=13 month=12 year=13 tag=rose
    """

    if not kwargs:
        return {"error": "You have given a condition"}

    return _atq(**kwargs)



"""
Tests for the spm repo
"""

import os
import shutil

import pytest
from tests.support.case import SPMCase


@pytest.mark.windows_whitelisted
@pytest.mark.destructive_test
class SPMRepoTest(SPMCase):
    """
    Validate commands related to spm repo
    """

    def setUp(self):
        self.config = self._spm_config()
        self._spm_build_files(self.config)

    def test_spm_create_update_repo(self):
        """
        test spm create_repo
        """
        self._spm_create_update_repo(self.config)

        self.assertTrue(os.path.exists(self.config["spm_db"]))

        l_repo_file = os.path.join(self.config["spm_cache_dir"], "local_repo.p")
        self.assertTrue(os.path.exists(l_repo_file))

    def tearDown(self):
        shutil.rmtree(self._tmp_spm)



"""
Test case for the vault execution module
"""


import salt.modules.vault as vault
from salt.exceptions import CommandExecutionError
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase


class TestVaultModule(LoaderModuleMockMixin, TestCase):
    """
    Test case for the vault execution module
    """

    def setup_loader_modules(self):
        return {
            vault: {
                "__opts__": {
                    "vault": {
                        "url": "http://127.0.0.1",
                        "auth": {"token": "test", "method": "token"},
                    }
                },
                "__grains__": {"id": "test-minion"},
            }
        }

    def test_read_secret_v1(self):
        """
        Test salt.modules.vault.read_secret function
        """
        version = {"v2": False, "data": None, "metadata": None, "type": None}
        mock_version = MagicMock(return_value=version)
        mock_vault = MagicMock()
        mock_vault.return_value.status_code = 200
        mock_vault.return_value.json.return_value = {"data": {"key": "test"}}
        with patch.dict(
            vault.__utils__, {"vault.make_request": mock_vault}
        ), patch.dict(vault.__utils__, {"vault.is_v2": mock_version}):
            vault_return = vault.read_secret("/secret/my/secret")

        self.assertDictEqual(vault_return, {"key": "test"})

    def test_read_secret_v1_key(self):
        """
        Test salt.modules.vault.read_secret function specifying key
        """
        version = {"v2": False, "data": None, "metadata": None, "type": None}
        mock_version = MagicMock(return_value=version)
        mock_vault = MagicMock()
        mock_vault.return_value.status_code = 200
        mock_vault.return_value.json.return_value = {"data": {"key": "somevalue"}}
        with patch.dict(
            vault.__utils__, {"vault.make_request": mock_vault}
        ), patch.dict(vault.__utils__, {"vault.is_v2": mock_version}):
            vault_return = vault.read_secret("/secret/my/secret", "key")

        self.assertEqual(vault_return, "somevalue")

    def test_read_secret_v2(self):
        """
        Test salt.modules.vault.read_secret function for v2 of kv secret backend
        """
        # given path secrets/mysecret generate v2 output
        version = {
            "v2": True,
            "data": "secrets/data/mysecret",
            "metadata": "secrets/metadata/mysecret",
            "type": "kv",
        }
        mock_version = MagicMock(return_value=version)
        mock_vault = MagicMock()
        mock_vault.return_value.status_code = 200
        v2_return = {
            "data": {
                "data": {"akey": "avalue"},
                "metadata": {
                    "created_time": "2018-10-23T20:21:55.042755098Z",
                    "destroyed": False,
                    "version": 13,
                    "deletion_time": "",
                },
            }
        }

        mock_vault.return_value.json.return_value = v2_return
        with patch.dict(
            vault.__utils__, {"vault.make_request": mock_vault}
        ), patch.dict(vault.__utils__, {"vault.is_v2": mock_version}):
            # Validate metadata returned
            vault_return = vault.read_secret("/secret/my/secret", metadata=True)
            self.assertDictContainsSubset({"data": {"akey": "avalue"}}, vault_return)
            # Validate just data returned
            vault_return = vault.read_secret("/secret/my/secret")
            self.assertDictContainsSubset({"akey": "avalue"}, vault_return)

    def test_read_secret_v2_key(self):
        """
        Test salt.modules.vault.read_secret function for v2 of kv secret backend
        with specified key
        """
        # given path secrets/mysecret generate v2 output
        version = {
            "v2": True,
            "data": "secrets/data/mysecret",
            "metadata": "secrets/metadata/mysecret",
            "type": "kv",
        }
        mock_version = MagicMock(return_value=version)
        mock_vault = MagicMock()
        mock_vault.return_value.status_code = 200
        v2_return = {
            "data": {
                "data": {"akey": "avalue"},
                "metadata": {
                    "created_time": "2018-10-23T20:21:55.042755098Z",
                    "destroyed": False,
                    "version": 13,
                    "deletion_time": "",
                },
            }
        }

        mock_vault.return_value.json.return_value = v2_return
        with patch.dict(
            vault.__utils__, {"vault.make_request": mock_vault}
        ), patch.dict(vault.__utils__, {"vault.is_v2": mock_version}):
            vault_return = vault.read_secret("/secret/my/secret", "akey")

        self.assertEqual(vault_return, "avalue")


class VaultDefaultTestCase(TestCase, LoaderModuleMockMixin):
    """
    Test cases for the default argument in the vault module

    NOTE: This test class is crafted such that the vault.make_request call will
    always fail. If you want to add other unit tests, you should put them in a
    separate class.
    """

    def setup_loader_modules(self):
        return {
            vault: {
                "__grains__": {"id": "foo"},
                "__utils__": {
                    "vault.make_request": MagicMock(side_effect=Exception("FAILED")),
                    "vault.is_v2": MagicMock(
                        return_value={
                            "v2": True,
                            "data": "secrets/data/mysecret",
                            "metadata": "secrets/metadata/mysecret",
                            "type": "kv",
                        }
                    ),
                },
            },
        }

    def setUp(self):
        self.path = "foo/bar/"

    def test_read_secret_with_default(self):
        assert vault.read_secret(self.path, default="baz") == "baz"

    def test_read_secret_no_default(self):
        try:
            vault.read_secret(self.path)
        except CommandExecutionError:
            # This is expected
            pass
        else:
            raise Exception("Should have raised a CommandExecutionError")

    def test_list_secrets_with_default(self):
        assert vault.list_secrets(self.path, default=["baz"]) == ["baz"]

    def test_list_secrets_no_default(self):
        try:
            vault.list_secrets(self.path)
        except CommandExecutionError:
            # This is expected
            pass
        else:
            raise Exception("Should have raised a CommandExecutionError")



"""
    Tests for salt.utils.boto3mod
"""

import random
import string

import salt.loader
import salt.utils.boto3mod as boto3mod
from salt.utils.versions import LooseVersion
from tests.support.mixins import LoaderModuleMockMixin
from tests.support.mock import MagicMock, patch
from tests.support.unit import TestCase, skipIf

try:
    import boto3
    from botocore.exceptions import ClientError

    HAS_BOTO3 = True
except ImportError:
    HAS_BOTO3 = False

REQUIRED_BOTO3_VERSION = "1.2.1"


@skipIf(HAS_BOTO3 is False, "The boto module must be installed.")
@skipIf(
    LooseVersion(boto3.__version__) < LooseVersion(REQUIRED_BOTO3_VERSION),
    "The boto3 module must be greater or equal to version {}".format(
        REQUIRED_BOTO3_VERSION
    ),
)
class Boto3modTestCase(TestCase, LoaderModuleMockMixin):
    """
    TestCase for salt.utils.boto3mod module
    """

    region = "us-east-1"
    service = "test-service"
    resource_name = "test-resource"
    resource_id = "test-resource-id"
    access_key = "GKTADJGHEIQSXMKKRBJ08H"
    secret_key = "askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs"
    conn_parameters = {}
    error_message = (
        "An error occurred ({}) when calling the {} operation: Test-defined error"
    )
    error_content = {"Error": {"Code": 101, "Message": "Test-defined error"}}
    session_ret = {}
    conn = None

    def setup_loader_modules(self):
        self.opts = {
            "__salt__": {"config.option": salt.config.DEFAULT_MINION_OPTS.copy()}
        }
        return {boto3mod: self.opts}

    def setUp(self):
        super().setUp()
        del self.opts

        # Set up MagicMock to replace the boto3 session
        # connections keep getting cached from prior tests, can't find the
        # correct context object to clear it. So randomize the cache key, to prevent any
        # cache hits
        self.conn_parameters = {
            "region": self.region,
            "keyid": self.secret_key,
            "profile": {},
        }
        self.conn_parameters["key"] = "".join(
            random.choice(string.ascii_lowercase + string.digits) for _ in range(50)
        )

        self.not_found_error = ClientError(
            {
                "Error": {
                    "Code": "ResourceNotFoundException",
                    "Message": "Test-defined error",
                }
            },
            "msg",
        )

        self.conn = MagicMock()
        self.addCleanup(delattr, self, "conn")
        self.patcher = patch("boto3.session.Session")
        self.addCleanup(self.patcher.stop)
        self.addCleanup(delattr, self, "patcher")
        mock_session = self.patcher.start()
        session_instance = mock_session.return_value
        session_instance.configure_mock(client=MagicMock(return_value=self.conn))
        self.paginator = MagicMock()
        self.addCleanup(delattr, self, "paginator")
        self.conn.configure_mock(get_paginator=MagicMock(return_value=self.paginator))

    def test_set_and_get_with_no_auth_params(self):
        boto3mod.cache_id(
            self.service, self.resource_name, resource_id=self.resource_id
        )
        self.assertEqual(
            boto3mod.cache_id(self.service, self.resource_name), self.resource_id
        )

    def test_set_and_get_with_explicit_auth_params(self):
        boto3mod.cache_id(
            self.service,
            self.resource_name,
            resource_id=self.resource_id,
            **self.conn_parameters
        )
        self.assertEqual(
            boto3mod.cache_id(self.service, self.resource_name, **self.conn_parameters),
            self.resource_id,
        )

    def test_set_and_get_with_different_region_returns_none(self):
        boto3mod.cache_id(
            self.service,
            self.resource_name,
            resource_id=self.resource_id,
            region="us-east-1",
        )
        self.assertEqual(
            boto3mod.cache_id(self.service, self.resource_name, region="us-west-2"),
            None,
        )

    def test_set_and_get_after_invalidation_returns_none(self):
        boto3mod.cache_id(
            self.service, self.resource_name, resource_id=self.resource_id
        )
        boto3mod.cache_id(
            self.service,
            self.resource_name,
            resource_id=self.resource_id,
            invalidate=True,
        )
        self.assertEqual(boto3mod.cache_id(self.service, self.resource_name), None)

    def test_partial(self):
        cache_id = boto3mod.cache_id_func(self.service)
        cache_id(self.resource_name, resource_id=self.resource_id)
        self.assertEqual(cache_id(self.resource_name), self.resource_id)



"""
Wrapper around Server Density API
=================================

.. versionadded:: 2014.7.0
"""


import logging
import os
import tempfile

import salt.utils.json
from salt.exceptions import CommandExecutionError

try:
    import requests

    ENABLED = True
except ImportError:
    ENABLED = False

log = logging.getLogger(__name__)


def __virtual__():
    """
    Return virtual name of the module.

    :return: The virtual name of the module.
    """
    if not ENABLED:
        return (False, "The requests python module cannot be imported")
    return "serverdensity_device"


def get_sd_auth(val, sd_auth_pillar_name="serverdensity"):
    """
    Returns requested Server Density authentication value from pillar.

    CLI Example:

    .. code-block:: bash

        salt '*' serverdensity_device.get_sd_auth <val>
    """
    sd_pillar = __pillar__.get(sd_auth_pillar_name)
    log.debug("Server Density Pillar: %s", sd_pillar)
    if not sd_pillar:
        log.error("Could not load %s pillar", sd_auth_pillar_name)
        raise CommandExecutionError(
            "{} pillar is required for authentication".format(sd_auth_pillar_name)
        )

    try:
        return sd_pillar[val]
    except KeyError:
        log.error("Could not find value %s in pillar", val)
        raise CommandExecutionError("{} value was not found in pillar".format(val))


def _clean_salt_variables(params, variable_prefix="__"):
    """
    Pops out variables from params which starts with `variable_prefix`.
    """
    list(list(map(params.pop, [k for k in params if k.startswith(variable_prefix)])))
    return params


def create(name, **params):
    """
    Function to create device in Server Density. For more info, see the `API
    docs`__.

    .. __: https://apidocs.serverdensity.com/Inventory/Devices/Creating

    CLI Example:

    .. code-block:: bash

        salt '*' serverdensity_device.create lama
        salt '*' serverdensity_device.create rich_lama group=lama_band installedRAM=32768
    """
    log.debug("Server Density params: %s", params)
    params = _clean_salt_variables(params)

    params["name"] = name
    api_response = requests.post(
        "https://api.serverdensity.io/inventory/devices/",
        params={"token": get_sd_auth("api_token")},
        data=params,
    )
    log.debug("Server Density API Response: %s", api_response)
    log.debug("Server Density API Response content: %s", api_response.content)
    if api_response.status_code == 200:
        try:
            return salt.utils.json.loads(api_response.content)
        except ValueError:
            log.error("Could not parse API Response content: %s", api_response.content)
            raise CommandExecutionError(
                "Failed to create, API Response: {}".format(api_response)
            )
    else:
        return None


def delete(device_id):
    """
    Delete a device from Server Density. For more information, see the `API
    docs`__.

    .. __: https://apidocs.serverdensity.com/Inventory/Devices/Deleting

    CLI Example:

    .. code-block:: bash

        salt '*' serverdensity_device.delete 51f7eafcdba4bb235e000ae4
    """
    api_response = requests.delete(
        "https://api.serverdensity.io/inventory/devices/" + device_id,
        params={"token": get_sd_auth("api_token")},
    )
    log.debug("Server Density API Response: %s", api_response)
    log.debug("Server Density API Response content: %s", api_response.content)
    if api_response.status_code == 200:
        try:
            return salt.utils.json.loads(api_response.content)
        except ValueError:
            log.error("Could not parse API Response content: %s", api_response.content)
            raise CommandExecutionError(
                "Failed to create, API Response: {}".format(api_response)
            )
    else:
        return None


def ls(**params):
    """
    List devices in Server Density

    Results will be filtered by any params passed to this function. For more
    information, see the API docs on listing_ and searching_.

    .. _listing: https://apidocs.serverdensity.com/Inventory/Devices/Listing
    .. _searching: https://apidocs.serverdensity.com/Inventory/Devices/Searching

    CLI Example:

    .. code-block:: bash

        salt '*' serverdensity_device.ls
        salt '*' serverdensity_device.ls name=lama
        salt '*' serverdensity_device.ls name=lama group=lama_band installedRAM=32768
    """
    params = _clean_salt_variables(params)

    endpoint = "devices"

    # Change endpoint if there are params to filter by:
    if params:
        endpoint = "resources"

    # Convert all ints to strings:
    for key, val in params.items():
        params[key] = str(val)

    api_response = requests.get(
        "https://api.serverdensity.io/inventory/{}".format(endpoint),
        params={
            "token": get_sd_auth("api_token"),
            "filter": salt.utils.json.dumps(params),
        },
    )
    log.debug("Server Density API Response: %s", api_response)
    log.debug("Server Density API Response content: %s", api_response.content)
    if api_response.status_code == 200:
        try:
            return salt.utils.json.loads(api_response.content)
        except ValueError:
            log.error(
                "Could not parse Server Density API Response content: %s",
                api_response.content,
            )
            raise CommandExecutionError(
                "Failed to create, Server Density API Response: {}".format(api_response)
            )
    else:
        return None


def update(device_id, **params):
    """
    Updates device information in Server Density. For more information see the
    `API docs`__.

    .. __: https://apidocs.serverdensity.com/Inventory/Devices/Updating

    CLI Example:

    .. code-block:: bash

        salt '*' serverdensity_device.update 51f7eafcdba4bb235e000ae4 name=lama group=lama_band
        salt '*' serverdensity_device.update 51f7eafcdba4bb235e000ae4 name=better_lama group=rock_lamas swapSpace=512
    """
    params = _clean_salt_variables(params)

    api_response = requests.put(
        "https://api.serverdensity.io/inventory/devices/" + device_id,
        params={"token": get_sd_auth("api_token")},
        data=params,
    )
    log.debug("Server Density API Response: %s", api_response)
    log.debug("Server Density API Response content: %s", api_response.content)
    if api_response.status_code == 200:
        try:
            return salt.utils.json.loads(api_response.content)
        except ValueError:
            log.error(
                "Could not parse Server Density API Response content: %s",
                api_response.content,
            )
            raise CommandExecutionError(
                "Failed to create, API Response: {}".format(api_response)
            )
    else:
        return None


def install_agent(agent_key, agent_version=1):
    """
    Function downloads Server Density installation agent, and installs sd-agent
    with agent_key. Optionally the agent_version would select the series to
    use (defaults on the v1 one).

    CLI Example:

    .. code-block:: bash

        salt '*' serverdensity_device.install_agent c2bbdd6689ff46282bdaa07555641498
        salt '*' serverdensity_device.install_agent c2bbdd6689ff46282bdaa07555641498 2
    """
    work_dir = os.path.join(__opts__["cachedir"], "tmp")
    if not os.path.isdir(work_dir):
        os.mkdir(work_dir)
    install_file = tempfile.NamedTemporaryFile(dir=work_dir, suffix=".sh", delete=False)
    install_filename = install_file.name
    install_file.close()

    account_field = "account_url"
    url = "https://www.serverdensity.com/downloads/agent-install.sh"
    if agent_version == 2:
        account_field = "account_name"
        url = "https://archive.serverdensity.com/agent-install.sh"

    account = get_sd_auth(account_field)

    __salt__["cmd.run"](
        cmd="curl -L {} -o {}".format(url, install_filename), cwd=work_dir
    )
    __salt__["cmd.run"](cmd="chmod +x {}".format(install_filename), cwd=work_dir)

    return __salt__["cmd.run"](
        cmd="{filename} -a {account} -k {agent_key}".format(
            filename=install_filename, account=account, agent_key=agent_key
        ),
        cwd=work_dir,
    )



"""
Unit Tests for functions located in salt.utils.doc.py.
"""


import salt.utils.doc
from tests.support.unit import TestCase


class DocUtilsTestCase(TestCase):
    """
    Test case for doc util.
    """

    def test_parse_docstring(self):
        test_keystone_str = """Management of Keystone users
                                ============================

                                :depends:   - keystoneclient Python module
                                :configuration: See :py:mod:`salt.modules.keystone` for setup instructions.
"""

        ret = salt.utils.doc.parse_docstring(test_keystone_str)
        expected_dict = {
            "deps": ["keystoneclient"],
            "full": (
                "Management of Keystone users\n                               "
                " ============================\n\n                               "
                " :depends:   - keystoneclient Python module\n                         "
                "       :configuration: See :py:mod:`salt.modules.keystone` for setup"
                " instructions.\n"
            ),
        }
        self.assertDictEqual(ret, expected_dict)



import salt.config
import salt.loader
from tests.support.unit import TestCase


class ConfirmTop(TestCase):
    def setUp(self):
        opts = salt.config.DEFAULT_MINION_OPTS.copy()
        self.matchers = salt.loader.matchers(opts)

    def test_sanity(self):
        match = self.matchers["confirm_top.confirm_top"]
        self.assertTrue(match("*", []))



"""
Return data by SMS.

.. versionadded:: 2015.5.0

:maintainer:    Damian Myerscough
:maturity:      new
:depends:       twilio
:platform:      all

To enable this returner the minion will need the python twilio library
installed and the following values configured in the minion or master
config:

.. code-block:: yaml

    twilio.sid: 'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX'
    twilio.token: 'XXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX'
    twilio.to: '+1415XXXXXXX'
    twilio.from: '+1650XXXXXXX'

To use the sms returner, append '--return sms' to the salt command.

.. code-block:: bash

    salt '*' test.ping --return sms

"""

import logging

import salt.returners

log = logging.getLogger(__name__)

try:
    import twilio

    # Grab version, ensure elements are ints
    twilio_version = tuple([int(x) for x in twilio.__version_info__])
    if twilio_version > (5,):
        TWILIO_5 = False
        from twilio.rest import Client as TwilioRestClient
        from twilio.rest import TwilioException as TwilioRestException
    else:
        TWILIO_5 = True
        from twilio.rest import TwilioRestClient
        from twilio import TwilioRestException  # pylint: disable=no-name-in-module

    HAS_TWILIO = True
except ImportError:
    HAS_TWILIO = False

__virtualname__ = "sms"


def __virtual__():
    if HAS_TWILIO:
        return __virtualname__

    return False, "Could not import sms returner; twilio is not installed."


def _get_options(ret=None):
    """
    Get the Twilio options from salt.
    """
    attrs = {"sid": "sid", "token": "token", "to": "to", "from": "from"}

    _options = salt.returners.get_returner_options(
        __virtualname__, ret, attrs, __salt__=__salt__, __opts__=__opts__
    )
    return _options


def returner(ret):
    """
    Return a response in an SMS message
    """
    _options = _get_options(ret)

    sid = _options.get("sid", None)
    token = _options.get("token", None)
    sender = _options.get("from", None)
    receiver = _options.get("to", None)

    if sid is None or token is None:
        log.error("Twilio sid/authentication token missing")
        return None

    if sender is None or receiver is None:
        log.error("Twilio to/from fields are missing")
        return None

    client = TwilioRestClient(sid, token)

    try:
        message = client.messages.create(
            body="Minion: {}\nCmd: {}\nSuccess: {}\n\nJid: {}".format(
                ret["id"], ret["fun"], ret["success"], ret["jid"]
            ),
            to=receiver,
            from_=sender,
        )
    except TwilioRestException as e:
        log.error("Twilio [https://www.twilio.com/docs/errors/%s]", e.code)
        return False

    return True



import pytest
import salt.utils.platform
from tests.support.case import ModuleCase
from tests.support.unit import skipIf


@skipIf(not salt.utils.platform.is_windows(), "windows test only")
@pytest.mark.windows_whitelisted
class WinServermanagerTest(ModuleCase):
    """
    Test for salt.modules.win_servermanager
    """

    @pytest.mark.slow_test
    def test_list_available(self):
        """
        Test list available features to install
        """
        cmd = self.run_function("win_servermanager.list_available")
        self.assertIn("DNS", cmd)
        self.assertIn("NetworkController", cmd)
        self.assertIn("RemoteAccess", cmd)



"""
Contains systemd related help files
"""

import logging
import os
import re
import subprocess

import salt.loader.context
import salt.utils.path
import salt.utils.stringutils
from salt.exceptions import SaltInvocationError

try:
    import dbus
except ImportError:
    dbus = None


log = logging.getLogger(__name__)


def booted(context=None):
    """
    Return True if the system was booted with systemd, False otherwise.  If the
    loader context dict ``__context__`` is passed, this function will set the
    ``salt.utils.systemd.booted`` key to represent if systemd is running and
    keep the logic below from needing to be run again during the same salt run.
    """
    contextkey = "salt.utils.systemd.booted"
    if isinstance(context, (dict, salt.loader.context.NamedLoaderContext)):
        # Can't put this if block on the same line as the above if block,
        # because it willl break the elif below.
        if contextkey in context:
            return context[contextkey]
    elif context is not None:
        raise SaltInvocationError("context must be a dictionary if passed")

    try:
        # This check does the same as sd_booted() from libsystemd-daemon:
        # http://www.freedesktop.org/software/systemd/man/sd_booted.html
        ret = bool(os.stat("/run/systemd/system"))
    except OSError:
        ret = False

    try:
        context[contextkey] = ret
    except TypeError:
        pass

    return ret


def offline(context=None):
    """Return True if systemd is in offline mode

    .. versionadded:: 3004
    """
    contextkey = "salt.utils.systemd.offline"
    if isinstance(context, (dict, salt.loader.context.NamedLoaderContext)):
        if contextkey in context:
            return context[contextkey]
    elif context is not None:
        raise SaltInvocationError("context must be a dictionary if passed")

    # Note that there is a difference from SYSTEMD_OFFLINE=1.  Here we
    # assume that there is no PID 1 to talk with.
    ret = not booted(context) and salt.utils.path.which("systemctl")

    try:
        context[contextkey] = ret
    except TypeError:
        pass

    return ret


def version(context=None):
    """
    Attempts to run systemctl --version. Returns None if unable to determine
    version.
    """
    contextkey = "salt.utils.systemd.version"
    if isinstance(context, (dict, salt.loader.context.NamedLoaderContext)):
        # Can't put this if block on the same line as the above if block,
        # because it will break the elif below.
        if contextkey in context:
            return context[contextkey]
    elif context is not None:
        raise SaltInvocationError("context must be a dictionary if passed")
    stdout = subprocess.Popen(
        ["systemctl", "--version"],
        close_fds=True,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
    ).communicate()[0]
    outstr = salt.utils.stringutils.to_str(stdout)
    try:
        ret = int(re.search(r"\w+ ([0-9]+)", outstr.splitlines()[0]).group(1))
    except (AttributeError, IndexError, ValueError):
        log.error(
            "Unable to determine systemd version from systemctl "
            "--version, output follows:\n%s",
            outstr,
        )
        return None
    else:
        try:
            context[contextkey] = ret
        except TypeError:
            pass
        return ret


def has_scope(context=None):
    """
    Scopes were introduced in systemd 205, this function returns a boolean
    which is true when the minion is systemd-booted and running systemd>=205.
    """
    if not booted(context):
        return False
    _sd_version = version(context)
    if _sd_version is None:
        return False
    return _sd_version >= 205


def pid_to_service(pid):
    """
    Check if a PID belongs to a systemd service and return its name.
    Return None if the PID does not belong to a service.

    Uses DBUS if available.
    """
    if dbus:
        return _pid_to_service_dbus(pid)
    else:
        return _pid_to_service_systemctl(pid)


def _pid_to_service_systemctl(pid):
    systemd_cmd = ["systemctl", "--output", "json", "status", str(pid)]
    try:
        systemd_output = subprocess.run(
            systemd_cmd, check=True, text=True, capture_output=True
        )
        status_json = salt.utils.json.find_json(systemd_output.stdout)
    except (ValueError, subprocess.CalledProcessError):
        return None

    name = status_json.get("_SYSTEMD_UNIT")
    if name and name.endswith(".service"):
        return _strip_suffix(name)
    else:
        return None


def _pid_to_service_dbus(pid):
    """
    Use DBUS to check if a PID belongs to a running systemd service and return the service name if it does.
    """
    bus = dbus.SystemBus()
    systemd_object = bus.get_object(
        "org.freedesktop.systemd1", "/org/freedesktop/systemd1"
    )
    systemd = dbus.Interface(systemd_object, "org.freedesktop.systemd1.Manager")
    try:
        service_path = systemd.GetUnitByPID(pid)
        service_object = bus.get_object("org.freedesktop.systemd1", service_path)
        service_props = dbus.Interface(
            service_object, "org.freedesktop.DBus.Properties"
        )
        service_name = service_props.Get("org.freedesktop.systemd1.Unit", "Id")
        name = str(service_name)

        if name and name.endswith(".service"):
            return _strip_suffix(name)
        else:
            return None
    except dbus.DBusException:
        return None


def _strip_suffix(service_name):
    """
    Strip ".service" suffix from a given service name.
    """
    return service_name[:-8]



"""
    :codeauthor: :email:`saltybob <bbaker@saltstack.com`
"""

from salt.client.ssh.wrapper import state
from tests.support.unit import TestCase


class StateTests(TestCase):
    def test_parse_mods(self):
        """
        Test _parse_mods
        """
        expected = ["a", "b", "c", "d", "e", "f"]
        mods = "a,b, c,  d,e ,f  "

        actual = state._parse_mods(mods)
        self.assertEqual(expected, actual)
